# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Relevance Pursuit model structure and optimization routines for the sparse optimization
of Gaussian process hyper-parameters, see [Ament2024pursuit]_ for details.

References

.. [Ament2024pursuit]
    S. Ament, E. Santorella, D. Eriksson, B. Letham, M. Balandat, and E. Bakshy.
    Robust Gaussian Processes via Relevance Pursuit. Advances in Neural Information
    Processing Systems 37, 2024. Arxiv: https://arxiv.org/abs/2410.24222.
"""

from __future__ import annotations

import math

from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from copy import copy, deepcopy
from functools import partial
from typing import Any, cast, Optional
from warnings import warn

import torch
from botorch.fit import fit_gpytorch_mll
from botorch.models.model import Model
from gpytorch.mlls.exact_marginal_log_likelihood import ExactMarginalLogLikelihood
from torch import Tensor
from torch.nn.parameter import Parameter

MLL_ITER = 10_000  # let's take convergence seriously
MLL_TOL = 1e-8
RESET_PARAMETERS = True
RESET_DENSE_PARAMETERS = False


class RelevancePursuitMixin(ABC):
    """Mixin class to convert between the sparse and dense representations of the
    relevance pursuit modules' sparse parameters, as well as to compute the generalized
    support acquisition and support deletion criteria.
    """

    dim: int  # the total number of features
    _support: list[int]  # indices of the features in the support, subset of range(dim)

    def __init__(
        self,
        dim: int,
        support: list[int] | None,
    ) -> None:
        """Constructor for the RelevancePursuitMixin class.

        For details, see [Ament2024pursuit]_ or https://arxiv.org/abs/2410.24222.

        Args:
            dim: The total number of features.
            support: The indices of the features in the support, subset of range(dim).
        """

        self.dim = dim
        self._support = support if support is not None else []
        # Assumption: sparse_parameter is initialized in sparse representation
        self._is_sparse = True
        self._expansion_modifier = None
        self._contraction_modifier = None

    @property
    @abstractmethod
    def sparse_parameter(self) -> Parameter:
        """The sparse parameter, required to have a single indexing dimension."""
        pass  # pragma: no cover

    @abstractmethod
    def set_sparse_parameter(self, value: Parameter) -> None:
        """Sets the sparse parameter.

        NOTE: We can't use the property setter @sparse_parameter.setter because of
        the special way PyTorch treats Parameter types, including custom setters that
        bypass the @property setters before the latter are called.
        """
        pass  # pragma: no cover

    @staticmethod
    def _from_model(model: Model) -> RelevancePursuitMixin:
        """Retrieves a RelevancePursuitMixin from a model."""
        raise NotImplementedError  # pragma: no cover

    @property
    def is_sparse(self) -> bool:
        # Do we need to differentiate between a full support sparse representation and
        # a full support dense representation? The order the of the indices could be
        # different, unless we keep them sorted.
        return self._is_sparse

    @property
    def support(self) -> list[int]:
        """The indices of the active parameters."""
        return self._support

    @property
    def is_active(self) -> Tensor:
        """A Boolean Tensor of length `dim`, indicating which of the `dim` indices of
        `self.sparse_parameter` are in the support, i.e. active."""
        is_active = [(i in self.support) for i in range(self.dim)]
        return torch.tensor(
            is_active, dtype=torch.bool, device=self.sparse_parameter.device
        )

    @property
    def inactive_indices(self) -> Tensor:
        """An integral Tensor of length `dim - len(support)`, indicating which of the
        indices of `self.sparse_parameter` are not in the support, i.e. inactive."""
        device = self.sparse_parameter.device
        return torch.arange(self.dim, device=device)[~self.is_active]

    def to_sparse(self) -> RelevancePursuitMixin:
        """Converts the sparse parameter to its sparse (< dim) representation.

        Returns:
            The current object in its sparse representation.
        """
        if not self.is_sparse:
            self.set_sparse_parameter(
                torch.nn.Parameter(self.sparse_parameter[self.support])
            )
            self._is_sparse = True
        return self

    def to_dense(self) -> RelevancePursuitMixin:
        """Converts the sparse parameter to its dense, length-`dim` representation.

        Returns:
            The current object in its dense representation.
        """
        if self.is_sparse:
            dtype = self.sparse_parameter.dtype
            device = self.sparse_parameter.device
            zero = torch.tensor(
                0.0,
                dtype=dtype,
                device=device,
            )
            dense_parameter = [
                (
                    self.sparse_parameter[self.support.index(i)]
                    if i in self.support
                    else zero
                )
                for i in range(self.dim)
            ]
            dense_parameter = torch.tensor(dense_parameter, dtype=dtype, device=device)
            self.set_sparse_parameter(torch.nn.Parameter(dense_parameter))
            self._is_sparse = False
        return self

    def expand_support(self, indices: list[int]) -> RelevancePursuitMixin:
        """Expands the support by a number of indices.

        Args:
            indices: A list of indices of `self.sparse_parameter` to add to the support.

        Returns:
            The current object, updated with the expanded support.
        """
        for i in indices:
            if i in self.support:
                raise ValueError(f"Feature {i} already in the support.")

        self.support.extend(indices)
        # we need to add the parameter in the sparse representation
        if self.is_sparse:
            self.set_sparse_parameter(
                torch.nn.Parameter(
                    torch.cat(
                        (
                            self.sparse_parameter,
                            torch.zeros(len(indices)).to(self.sparse_parameter),
                        )
                    )
                )
            )
        return self

    def contract_support(self, indices: list[int]) -> RelevancePursuitMixin:
        """Contracts the support by a number of indices.

        Args:
            indices: A list of indices of `self.sparse_parameter` to remove from
                the support.

        Returns:
            The current object, updated with the contracted support.
        """
        # indices into the sparse representation of features to *keep*
        sparse_indices = list(range(len(self.support)))
        original_support = copy(self.support)
        for i in indices:
            if i not in self.support:
                raise ValueError(f"Feature {i} is not in support.")
            sparse_indices.remove(original_support.index(i))
            self.support.remove(i)

        # we need to remove the parameter in the sparse representation
        if self.is_sparse:
            self.set_sparse_parameter(Parameter(self.sparse_parameter[sparse_indices]))
        else:
            requires_grad = self.sparse_parameter.requires_grad
            self.sparse_parameter.requires_grad_(False)
            self.sparse_parameter[indices] = 0.0
            self.sparse_parameter.requires_grad_(requires_grad)  # restore
        return self

    # support initialization helpers
    def full_support(self) -> RelevancePursuitMixin:
        """Initializes the RelevancePursuitMixin with a full, size-`dim` support.

        Returns:
            The current object with full support in the dense representation.
        """
        self.expand_support([i for i in range(self.dim) if i not in self.support])
        self.to_dense()  # no reason to be sparse with full support
        return self

    def remove_support(self) -> RelevancePursuitMixin:
        """Initializes the RelevancePursuitMixin with an empty, size-zero support.

        Returns:
            The current object with empty support, representation unchanged.
        """
        self._support = []
        requires_grad = self.sparse_parameter.requires_grad
        if self.is_sparse:
            self.set_sparse_parameter(
                torch.nn.Parameter(torch.tensor([]).to(self.sparse_parameter))
            )
        else:
            self.sparse_parameter.requires_grad_(False)
            self.sparse_parameter[:] = 0.0
        self.sparse_parameter.requires_grad_(requires_grad)
        return self

    # the following two methods are the only ones that are specific to the marginal
    # likelihood optimization problem
    def support_expansion(
        self,
        mll: ExactMarginalLogLikelihood,
        n: int = 1,
        modifier: Callable[[Tensor], Tensor] | None = None,
    ) -> bool:
        """Computes the indices of the elements that maximize the gradient of the sparse
        parameter and that are not already in the support, and subsequently expands the
        support to include the elements if their gradient is positive.

        Args:
            mll: The marginal likelihood, containing the model to optimize.
                NOTE: Virtually all of the rest of the code is not specific to the
                marginal likelihood optimization, so we could generalize this to work
                with any objective.
            n: The maximum number of elements to select. NOTE: The actual number of
                elements that are added could be fewer if there are fewer than `n`
                elements with a positive gradient.
            modifier: A function that modifies the gradient of the inactive elements
                before computing the support expansion criterion. This can be used
                to select the maximum gradient magnitude for real-valued elements
                whose gradients are not non-negative, using modifier = torch.abs.

        Returns:
            True if the support was expanded, False otherwise.
        """
        # can't expand if the support is already full, or if n is non-positive
        if len(self.support) == self.dim or n <= 0:
            return False

        g = self.expansion_objective(mll)

        modifier = modifier if modifier is not None else self._expansion_modifier
        if modifier is not None:
            g = modifier(g)

        # support is already removed from consideration
        # gradient of the support parameters is not necessarily zero,
        # even for a converged solution in the presence of constraints.
        # NOTE: these indices are relative to self.inactive_indices.
        indices = g.argsort(descending=True)[:n]
        indices = indices[g[indices] > 0]
        if indices.numel() == 0:  # no indices with positive gradient
            return False
        self.expand_support(self.inactive_indices[indices].tolist())
        return True

    def expansion_objective(self, mll: ExactMarginalLogLikelihood) -> Tensor:
        """Computes an objective value for all the inactive parameters, i.e.
        self.sparse_parameter[~self.is_active] since we can't add already active
        parameters to the support. This value will be used to select the parameters.

        Args:
            mll: The marginal likelihood, containing the model to optimize.

        Returns:
            The expansion objective value for all the inactive parameters.
        """
        return self._sparse_parameter_gradient(mll)

    def _sparse_parameter_gradient(self, mll: ExactMarginalLogLikelihood) -> Tensor:
        """Computes the gradient of the marginal likelihood with respect to the
        sparse parameter.

        Args:
            mll: The marginal likelihood, containing the model to optimize.

        Returns:
            The gradient of the marginal likelihood with respect to the inactive
            sparse parameters.
        """
        # evaluate gradient of the sparse parameter
        is_sparse = self.is_sparse  # in order to restore the original representation
        self.to_dense()  # need the parameter in its dense parameterization

        requires_grad = self.sparse_parameter.requires_grad
        self.sparse_parameter.requires_grad_(True)
        if self.sparse_parameter.grad is not None:
            self.sparse_parameter.grad.zero_()
        mll.train()  # NOTE: this changes model.train_inputs
        model = mll.model
        X, Y = model.train_inputs[0], model.train_targets
        cast(
            Tensor,
            mll(
                mll.model(X),
                Y,
                *(model.transform_inputs(X=t_in) for t_in in model.train_inputs),
            ),
        ).backward()  # evaluation
        self.sparse_parameter.requires_grad_(requires_grad)

        g = self.sparse_parameter.grad
        if g is None:
            raise ValueError(
                "The gradient of the sparse_parameter is None, most likely "
                "because the passed marginal likelihood is not a function of the "
                "sparse_parameter."
            )

        if is_sparse:
            self.to_sparse()

        return g[~self.is_active]  # only need the inactive parameters

    def support_contraction(
        self,
        mll: ExactMarginalLogLikelihood,
        n: int = 1,
        modifier: Callable[[Tensor], Tensor] | None = None,
    ) -> bool:
        """Computes the indices of the elements with the smallest magnitude,
        and subsequently contracts the support by exluding the elements.

        Args:
            mll: The marginal likelihood, containing the model to optimize.
                NOTE: Virtually all of the rest of the code is not specific to the
                marginal likelihood optimization, so we could generalize this to work
                with any objective.
            n: The number of elements to select for removal.
            modifier: A function that modifies the parameter values before computing
                the support contraction criterion.

        Returns:
            True if the support was expanded, False otherwise.
        """
        # can't expand if the support is already empty, or if n is non-positive
        if len(self.support) == 0 or n <= 0:
            return False

        is_sparse = self.is_sparse
        self.to_sparse()
        x = self.sparse_parameter

        modifier = modifier if modifier is not None else self._contraction_modifier
        if modifier is not None:
            x = modifier(x)

        # for non-negative parameters, could break ties at zero
        # based on derivative
        sparse_indices = x.argsort(descending=False)[:n]
        indices = [self.support[i] for i in sparse_indices]
        self.contract_support(indices)
        if not is_sparse:
            self.to_dense()
        return True

    def optimize_mll(
        self,
        mll: ExactMarginalLogLikelihood,
        model_trace: list[Model] | None = None,
        reset_parameters: bool = RESET_PARAMETERS,
        reset_dense_parameters: bool = RESET_DENSE_PARAMETERS,
        # fit_gpytorch_mll kwargs
        closure: Callable[[], tuple[Tensor, Sequence[Tensor | None]]] | None = None,
        optimizer: Callable | None = None,
        closure_kwargs: dict[str, Any] | None = None,
        optimizer_kwargs: dict[str, Any] | None = None,
    ):
        """Optimizes the marginal likelihood.

        Args:
            mll: The marginal likelihood, containing the model to optimize.
            model_trace: If not None, a list to which a deepcopy of the model state is
                appended. NOTE This operation is *in place*.
            reset_parameters: If True, initializes the sparse parameter to the all-zeros
                vector before every marginal likelihood optimization step. If False, the
                optimization is warm-started with the previous iteration's parameters.
            reset_dense_parameters: If True, re-initializes the dense parameters, e.g.
                other GP hyper-parameters that are *not* part of the Relevance Pursuit
                module, to the initial values provided by their associated constraints.
            closure: A closure to use to compute the loss and the gradients, see
                docstring of `fit_gpytorch_mll` for details.
            optimizer: The numerical optimizer, see docstring of `fit_gpytorch_mll`.
            closure_kwargs: Additional arguments to pass to the `closure` function.
            optimizer_kwargs: A dictionary of keyword arguments for the optimizer.

        Returns:
            The marginal likelihood after optimization.
        """
        if reset_parameters:
            # this might be beneficial because the parameters can
            # end up at a constraint boundary, which can anecdotally make
            # it more difficult to move the newly added parameters.
            with torch.no_grad():
                self.sparse_parameter.zero_()

        if reset_dense_parameters:
            initialize_dense_parameters(mll.model)

        # move to sparse representation for optimization
        # NOTE: this function should never force the dense representation, because some
        # models might never need it, and it would be inefficient.
        self.to_sparse()
        mll = fit_gpytorch_mll(
            mll,
            optimizer_kwargs=optimizer_kwargs,
            closure=closure,
            optimizer=optimizer,
            closure_kwargs=closure_kwargs,
        )
        if model_trace is not None:
            # need to record the full model here, rather than just the sparse parameter
            # since other hyper-parameters are co-adapted to the sparse parameter.
            model_trace.append(deepcopy(mll.model))
        return mll


# Optimization Algorithms
def forward_relevance_pursuit(
    sparse_module: RelevancePursuitMixin,
    mll: ExactMarginalLogLikelihood,
    sparsity_levels: list[int] | None = None,
    reset_parameters: bool = RESET_PARAMETERS,
    reset_dense_parameters: bool = RESET_DENSE_PARAMETERS,
    record_model_trace: bool = True,
    initial_support: list[int] | None = None,
    # fit_gpytorch_mll kwargs
    closure: Callable[[], tuple[Tensor, Sequence[Tensor | None]]] | None = None,
    optimizer: Callable | None = None,
    closure_kwargs: dict[str, Any] | None = None,
    optimizer_kwargs: dict[str, Any] | None = None,
) -> tuple[RelevancePursuitMixin, Optional[list[Model]]]:
    """Forward Relevance Pursuit.

    NOTE: For the robust `SparseOutlierNoise` model of [Ament2024pursuit]_, the forward
    algorithm is generally faster than the backward algorithm, particularly when the
    maximum sparsity level is small, but it leads to less robust results when the number
    of outliers is large.

    For details, see [Ament2024pursuit]_ or https://arxiv.org/abs/2410.24222.

    Example:
        >>> base_noise = HomoskedasticNoise(
        >>>    noise_constraint=NonTransformedInterval(
        >>>        1e-5, 1e-1, initial_value=1e-3
        >>>    )
        >>> )
        >>> likelihood = SparseOutlierGaussianLikelihood(
        >>>    base_noise=base_noise,
        >>>    dim=X.shape[0],
        >>> )
        >>> model = SingleTaskGP(train_X=X, train_Y=Y, likelihood=likelihood)
        >>> mll = ExactMarginalLogLikelihood(model.likelihood, model)
        >>> # NOTE: `likelihood.noise_covar` is the `RelevancePursuitMixin`
        >>> sparse_module = likelihood.noise_covar
        >>> sparse_module, model_trace = forward_relevance_pursuit(sparse_module, mll)

    Args:
        sparse_module: The relevance pursuit module.
        mll: The marginal likelihood, containing the model to optimize.
        sparsity_levels: The sparsity levels to expand the support to.
        reset_parameters: If true, initializes the sparse parameter to the all zeros
            after each iteration.
        reset_dense_parameters: If true, re-initializes the dense parameters, e.g.
            other GP hyper-parameters that are *not* part of the Relevance Pursuit
            module, to the initial values provided by their associated constraints.
        record_model_trace: If true, records the model state after every iteration.
        initial_support: The support with which to initialize the sparse module. By
            default, the support is initialized to the empty set.
        closure: A closure to use to compute the loss and the gradients, see docstring
            of `fit_gpytorch_mll` for details.
        optimizer: The numerical optimizer, see docstring of `fit_gpytorch_mll`.
        closure_kwargs: Additional arguments to pass to the `closure` function.
        optimizer_kwargs: A dictionary of keyword arguments to pass to the optimizer.
            By default, initializes the "options" sub-dictionary with `maxiter` and
            `ftol`, `gtol` values, unless specified.

    Returns:
        The relevance pursuit module after forward relevance pursuit optimization, and
        a list of models with different supports that were optimized.
    """
    sparse_module.remove_support()
    if initial_support is not None:
        sparse_module.expand_support(initial_support)

    if sparsity_levels is None:
        sparsity_levels = list(range(len(sparse_module.support), sparse_module.dim + 1))

    # since this is the forward algorithm, potential sparsity levels
    # must be in increasing order and unique.
    sparsity_levels = list(set(sparsity_levels))
    sparsity_levels.sort(reverse=False)

    optimizer_kwargs = _initialize_optimizer_kwargs(optimizer_kwargs)

    model_trace = [] if record_model_trace else None

    optimize_mll = partial(
        sparse_module.optimize_mll,
        model_trace=model_trace,
        reset_parameters=reset_parameters,
        reset_dense_parameters=reset_dense_parameters,
        # These are the args of the canonical mll fit routine
        closure=closure,
        optimizer=optimizer,
        closure_kwargs=closure_kwargs,
        optimizer_kwargs=optimizer_kwargs,
    )

    # if sparsity levels contains the initial support, remove it
    if sparsity_levels[0] == len(sparse_module.support):
        sparsity_levels.pop(0)

    optimize_mll(mll)  # initial optimization

    for sparsity in sparsity_levels:
        support_size = len(sparse_module.support)
        num_expand = sparsity - support_size
        expanded = sparse_module.support_expansion(mll=mll, n=num_expand)
        if not expanded:  # stationary support
            warn(
                "Terminating optimization because the expansion from sparsity "
                f"{support_size} to {sparsity} was unsuccessful, usually due to "
                "reaching a stationary point of the marginal likelihood.",
                Warning,
                stacklevel=2,
            )
            break

        optimize_mll(mll)  # re-optimize support

    return sparse_module, model_trace


def backward_relevance_pursuit(
    sparse_module: RelevancePursuitMixin,
    mll: ExactMarginalLogLikelihood,
    sparsity_levels: list[int] | None = None,
    reset_parameters: bool = RESET_PARAMETERS,
    reset_dense_parameters: bool = RESET_DENSE_PARAMETERS,
    record_model_trace: bool = True,
    initial_support: list[int] | None = None,
    # fit_gpytorch_mll kwargs
    closure: Callable[[], tuple[Tensor, Sequence[Tensor | None]]] | None = None,
    optimizer: Callable | None = None,
    closure_kwargs: dict[str, Any] | None = None,
    optimizer_kwargs: dict[str, Any] | None = None,
) -> tuple[RelevancePursuitMixin, Optional[list[Model]]]:
    """Backward Relevance Pursuit.

    NOTE: For the robust `SparseOutlierNoise` model of [Ament2024pursuit]_, the backward
    algorithm generally leads to more robust results than the forward algorithm,
    especially when the number of outliers is large, but is more expensive unless the
    support is contracted by more than one in each iteration.

    For details, see [Ament2024pursuit]_ or https://arxiv.org/abs/2410.24222.

    Example:
        >>> base_noise = HomoskedasticNoise(
        >>>    noise_constraint=NonTransformedInterval(
        >>>        1e-5, 1e-1, initial_value=1e-3
        >>>    )
        >>> )
        >>> likelihood = SparseOutlierGaussianLikelihood(
        >>>    base_noise=base_noise,
        >>>    dim=X.shape[0],
        >>> )
        >>> model = SingleTaskGP(train_X=X, train_Y=Y, likelihood=likelihood)
        >>> mll = ExactMarginalLogLikelihood(model.likelihood, model)
        >>> # NOTE: `likelihood.noise_covar` is the `RelevancePursuitMixin`
        >>> sparse_module = likelihood.noise_covar
        >>> sparse_module, model_trace = backward_relevance_pursuit(sparse_module, mll)

    Args:
        sparse_module: The relevance pursuit module.
        mll: The marginal likelihood, containing the model to optimize.
        sparsity_levels: The sparsity levels to expand the support to.
        reset_parameters: If true, initializes the sparse parameter to the all zeros
            after each iteration.
        reset_dense_parameters: If true, re-initializes the dense parameters, e.g.
            other GP hyper-parameters that are *not* part of the Relevance Pursuit
            module, to the initial values provided by their associated constraints.
        record_model_trace: If true, records the model state after every iteration.
        initial_support: The support with which to initialize the sparse module. By
            default, the support is initialized to the full set.
        closure: A closure to use to compute the loss and the gradients, see docstring
            of `fit_gpytorch_mll` for details.
        optimizer: The numerical optimizer, see docstring of `fit_gpytorch_mll`.
        closure_kwargs: Additional arguments to pass to the `closure` function.
        optimizer_kwargs: A dictionary of keyword arguments to pass to the optimizer.
            By default, initializes the "options" sub-dictionary with `maxiter` and
            `ftol`, `gtol` values, unless specified.

    Returns:
        The relevance pursuit module after forward relevance pursuit optimization, and
        a list of models with different supports that were optimized.
    """
    if initial_support is not None:
        sparse_module.remove_support()
        sparse_module.expand_support(initial_support)
    else:
        sparse_module.full_support()

    if sparsity_levels is None:
        sparsity_levels = list(range(len(sparse_module.support) + 1))

    # since this is the backward algorithm, potential sparsity levels
    # must be in decreasing order, unique, and less than the initial support.
    sparsity_levels = list(set(sparsity_levels))
    sparsity_levels.sort(reverse=True)

    optimizer_kwargs = _initialize_optimizer_kwargs(optimizer_kwargs)

    model_trace = [] if record_model_trace else None

    def optimize_mll(mll):
        return sparse_module.optimize_mll(
            mll=mll,
            model_trace=model_trace,
            reset_parameters=reset_parameters,
            reset_dense_parameters=reset_dense_parameters,
            # These are the args of the canonical mll fit routine
            closure=closure,
            optimizer=optimizer,
            closure_kwargs=closure_kwargs,
            optimizer_kwargs=optimizer_kwargs,
        )

    # if sparsity levels contains the initial support, remove it
    if sparsity_levels[0] == len(sparse_module.support):
        sparsity_levels.pop(0)

    optimize_mll(mll)  # initial optimization

    for sparsity in sparsity_levels:
        support_size = len(sparse_module.support)
        num_contract = support_size - sparsity
        contracted = sparse_module.support_contraction(mll=mll, n=num_contract)
        if not contracted:  # stationary support
            warn(
                "Terminating optimization because the contraction from sparsity "
                f"{support_size} to {sparsity} was unsuccessful.",
                Warning,
                stacklevel=2,
            )
            break

        optimize_mll(mll)  # re-optimize support

    return sparse_module, model_trace


# Bayesian Model Comparison
def get_posterior_over_support(
    rp_class: type[RelevancePursuitMixin],
    model_trace: list[Model],
    log_support_prior: Callable[[Tensor], Tensor] | None = None,
    prior_mean_of_support: float | None = None,
) -> tuple[Tensor, Tensor]:
    """Computes the posterior distribution over a list of models.
    Assumes we are storing both likelihood and GP model in the model_trace.

    Example:
        >>> likelihood = SparseOutlierGaussianLikelihood(
        >>>    base_noise=base_noise,
        >>>    dim=X.shape[0],
        >>> )
        >>> model = SingleTaskGP(train_X=X, train_Y=Y, likelihood=likelihood)
        >>> mll = ExactMarginalLogLikelihood(model.likelihood, model)
        >>> # NOTE: `likelihood.noise_covar` is the `RelevancePursuitMixin`
        >>> sparse_module = likelihood.noise_covar
        >>> sparse_module, model_trace = backward_relevance_pursuit(sparse_module, mll)
        >>> # NOTE: SparseOutlierNoise is the type of `sparse_module`
        >>> support_size, bmc_probabilities = get_posterior_over_support(
        >>>    SparseOutlierNoise, model_trace, prior_mean_of_support=2.0
        >>> )

    Args:
        rp_class: The relevance pursuit class to use for computing the support size.
            This is used to get the RelevancePursuitMixin from the Model via the static
            method `_from_model`. We could generalize this and let the user pass this
            getter instead.
        model_trace: A list of models with different support sizes, usually generated
            with relevance_pursuit.
        log_support_prior: Callable that computes the log prior probability of a
            support size. If None, uses a default exponential prior with a mean
            specified by `prior_mean_of_support`.
        prior_mean_of_support: A mean value for the default exponential prior
            distribution over the support size. Ignored if `log_support_prior`
            is passed.

    Returns:
        A tensor of posterior marginal likelihoods, one for each model in the trace.
    """
    if log_support_prior is None:
        if prior_mean_of_support is None:
            raise ValueError(
                "`log_support_prior` and `prior_mean_of_support` cannot both be None."
            )
        log_support_prior = partial(_exp_log_pdf, mean=prior_mean_of_support)

    log_support_prior = cast(Callable[[Tensor], Tensor], log_support_prior)

    def log_prior(
        model: Model,
        dtype: torch.dtype,
        device: torch.device,
    ) -> tuple[Tensor, Tensor]:
        sparse_module = rp_class._from_model(model)
        num_support = torch.tensor(
            len(sparse_module.support), dtype=dtype, device=device
        )
        return num_support, log_support_prior(num_support)  # pyre-ignore[29]

    log_mll_trace = []
    log_prior_trace = []
    support_size_trace = []
    for model in model_trace:
        mll = ExactMarginalLogLikelihood(likelihood=model.likelihood, model=model)
        mll.train()
        X, Y = mll.model.train_inputs[0], mll.model.train_targets
        F = mll.model(X)
        TX = mll.model.transform_inputs(X) if mll.model.training else X
        mll_i = cast(Tensor, mll(F, Y, TX))
        log_mll_trace.append(mll_i)
        support_size, log_prior_i = log_prior(
            model,
            dtype=mll_i.dtype,
            device=mll_i.device,
        )
        support_size_trace.append(support_size)
        log_prior_trace.append(log_prior_i)

    log_mll_trace = torch.stack(log_mll_trace)
    log_prior_trace = torch.stack(log_prior_trace)
    support_size_trace = torch.stack(support_size_trace)

    unnormalized_posterior_trace = log_mll_trace + log_prior_trace
    evidence = unnormalized_posterior_trace.logsumexp(dim=-1)
    posterior_probabilities = (unnormalized_posterior_trace - evidence).exp()
    return support_size_trace, posterior_probabilities


def _exp_log_pdf(x: Tensor, mean: Tensor) -> Tensor:
    """Compute the exponential log probability density.

    Args:
        x: A tensor of values.
        mean: A tensor of means.

    Returns:
        A tensor of log probabilities.
    """
    return -x / mean - math.log(mean)


def initialize_dense_parameters(model: Model) -> tuple[Model, dict[str, Any]]:
    """Sets the dense parameters of a model to their initial values. Infers initial
    values from the constraints, if no initial values are provided. If a parameter
    does not have a constraint, it is initialized to zero.

    Args:
        model: The model to initialize.

    Returns:
        The re-initialized model, and a dictionary of initial values.
    """
    constraints = dict(model.named_constraints())
    parameters = dict(model.named_parameters())
    initial_values = {
        n: getattr(constraints.get(n + "_constraint", None), "_initial_value", None)
        for n in parameters
    }
    lower_bounds = {
        n: getattr(
            constraints.get(n + "_constraint", None),
            "lower_bound",
            torch.tensor(-torch.inf),
        )
        for n in parameters
    }
    upper_bounds = {
        n: getattr(
            constraints.get(n + "_constraint", None),
            "upper_bound",
            torch.tensor(torch.inf),
        )
        for n in parameters
    }
    for name, value in initial_values.items():
        initial_values[name] = _get_initial_value(
            value=value,
            lower=lower_bounds[name],
            upper=upper_bounds[name],
        )

    # the initial values need to be converted to the transformed space
    for n, v in initial_values.items():
        c = constraints.get(n + "_constraint", None)
        # convert the constraint into the latent space
        if c is not None:
            initial_values[n] = c.inverse_transform(v)
    model.initialize(**initial_values)
    parameters = dict(model.named_parameters())
    return model, initial_values


def _get_initial_value(value: Tensor, lower: Tensor, upper: Tensor) -> Tensor:
    # if no initial value is provided, or the initial value is outside the bounds,
    # use a rule-based initialization.
    if value is None or not ((lower <= value) and (value <= upper)):
        if upper.isinf():
            value = 0.0 if lower.isinf() else lower + 1
        elif lower.isinf():  # implies u[n] is finite
            value = upper - 1
        else:  # both are finite
            # generally keep the value close to the lower bound in this case,
            # since many parameters (e.g. lengthscales) exhibit vanishing gradients
            # for large values.
            value = lower + torch.minimum(
                torch.ones_like(lower),
                (upper - lower) / 2,
            )
    return torch.as_tensor(value, dtype=lower.dtype, device=lower.device)


def _initialize_optimizer_kwargs(
    optimizer_kwargs: dict[str, Any] | None,
) -> dict[str, Any]:
    """Initializes the optimizer kwargs with default values if they are not provided.

    Args:
        optimizer_kwargs: The optimizer kwargs to initialize.

    Returns:
        The initialized optimizer kwargs.
    """
    if optimizer_kwargs is None:
        optimizer_kwargs = {}

    if optimizer_kwargs.get("options") is None:
        optimizer_kwargs["options"] = {}

    options = optimizer_kwargs["options"]
    if "maxiter" not in options:
        options.update({"maxiter": MLL_ITER})

    if ("ftol" not in options) and ("gtol" not in options):
        options.update({"ftol": MLL_TOL})
        options.update({"gtol": MLL_TOL})

    return optimizer_kwargs



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""Hypervolume Utilities.

References

.. [Fonseca2006]
    C. M. Fonseca, L. Paquete, and M. Lopez-Ibanez. An improved dimension-sweep
    algorithm for the hypervolume indicator. In IEEE Congress on Evolutionary
    Computation, pages 1157-1163, Vancouver, Canada, July 2006.

.. [Ishibuchi2011]
    H. Ishibuchi, N. Akedo, and Y. Nojima. A many-objective test problem
    for visually examining diversity maintenance behavior in a decision
    space. Proc. 13th Annual Conf. Genetic Evol. Comput., 2011.
"""

from __future__ import annotations

import random

import warnings
from collections.abc import Callable
from copy import deepcopy

from itertools import combinations

import torch
from botorch.acquisition.cached_cholesky import CachedCholeskyMCSamplerMixin
from botorch.acquisition.multi_objective.objective import MCMultiOutputObjective
from botorch.acquisition.multi_objective.utils import (
    prune_inferior_points_multi_objective,
)
from botorch.exceptions.errors import (
    BotorchError,
    BotorchTensorDimensionError,
    UnsupportedError,
)
from botorch.exceptions.warnings import BotorchWarning
from botorch.models.model import Model
from botorch.sampling.base import MCSampler
from botorch.utils.multi_objective.box_decompositions.box_decomposition_list import (
    BoxDecompositionList,
)
from botorch.utils.multi_objective.box_decompositions.dominated import (
    DominatedPartitioning,
)
from botorch.utils.multi_objective.box_decompositions.non_dominated import (
    FastNondominatedPartitioning,
    NondominatedPartitioning,
)
from botorch.utils.multi_objective.box_decompositions.utils import (
    _pad_batch_pareto_frontier,
)
from botorch.utils.objective import compute_feasibility_indicator
from botorch.utils.torch import BufferDict
from torch import Tensor

MIN_Y_RANGE = 1e-7


def infer_reference_point(
    pareto_Y: Tensor,
    max_ref_point: Tensor | None = None,
    scale: float = 0.1,
    scale_max_ref_point: bool = False,
) -> Tensor:
    r"""Get reference point for hypervolume computations.

    This sets the reference point to be `ref_point = nadir - scale * range`
    when there is no `pareto_Y` that is better than `max_ref_point`.
    If there's `pareto_Y` better than `max_ref_point`, the reference point
    will be set to `max_ref_point - scale * range` if `scale_max_ref_point`
    is true and to `max_ref_point` otherwise.

    [Ishibuchi2011]_ find 0.1 to be a robust multiplier for scaling the
    nadir point.

    Note: this assumes maximization of all objectives.

    Args:
        pareto_Y: A `n x m`-dim tensor of Pareto-optimal points.
        max_ref_point: A `m` dim tensor indicating the maximum reference point.
            Some elements can be NaN, except when `pareto_Y` is empty,
            in which case these dimensions will be treated as if no
            `max_ref_point` was provided and set to `nadir - scale * range`.
        scale: A multiplier used to scale back the reference point based on the
            range of each objective.
        scale_max_ref_point: A boolean indicating whether to apply scaling to
            the max_ref_point based on the range of each objective.

    Returns:
        A `m`-dim tensor containing the reference point.
    """
    if pareto_Y.shape[0] == 0:
        if max_ref_point is None:
            raise BotorchError("Empty pareto set and no max ref point provided")
        if max_ref_point.isnan().any():
            raise BotorchError("Empty pareto set and max ref point includes NaN.")
        if scale_max_ref_point:
            return max_ref_point - scale * max_ref_point.abs()
        return max_ref_point
    if max_ref_point is not None:
        non_nan_idx = ~max_ref_point.isnan()
        # Count all points exceeding non-NaN reference point as being better.
        better_than_ref = (pareto_Y[:, non_nan_idx] > max_ref_point[non_nan_idx]).all(
            dim=-1
        )
    else:
        non_nan_idx = torch.ones(
            pareto_Y.shape[-1], dtype=torch.bool, device=pareto_Y.device
        )
        better_than_ref = torch.ones(
            pareto_Y.shape[:1], dtype=torch.bool, device=pareto_Y.device
        )
    if max_ref_point is not None and better_than_ref.any() and non_nan_idx.all():
        Y_range = pareto_Y[better_than_ref].max(dim=0).values - max_ref_point
        if scale_max_ref_point:
            return max_ref_point - scale * Y_range
        return max_ref_point
    elif pareto_Y.shape[0] == 1:
        # no points better than max_ref_point and only a single observation
        # subtract MIN_Y_RANGE to handle the case that pareto_Y is a singleton
        # with objective value of 0.
        Y_range = pareto_Y.abs().clamp_min(MIN_Y_RANGE).view(-1)
        ref_point = pareto_Y.view(-1) - scale * Y_range
    else:
        # no points better than max_ref_point and multiple observations
        # make sure that each dimension of the nadir point is no greater than
        # the max_ref_point
        nadir = pareto_Y.min(dim=0).values
        if max_ref_point is not None:
            nadir[non_nan_idx] = torch.min(
                nadir[non_nan_idx], max_ref_point[non_nan_idx]
            )
        ideal = pareto_Y.max(dim=0).values
        # handle case where all values for one objective are the same
        Y_range = (ideal - nadir).clamp_min(MIN_Y_RANGE)
        ref_point = nadir - scale * Y_range
    # Set not-nan indices - if any - to max_ref_point.
    if non_nan_idx.any() and not non_nan_idx.all() and better_than_ref.any():
        if scale_max_ref_point:
            ref_point[non_nan_idx] = (max_ref_point - scale * Y_range)[non_nan_idx]
        else:
            ref_point[non_nan_idx] = max_ref_point[non_nan_idx]
    return ref_point


class Hypervolume:
    r"""Hypervolume computation dimension sweep algorithm from [Fonseca2006]_.

    Adapted from Simon Wessing's implementation of the algorithm
    (Variant 3, Version 1.2) in [Fonseca2006]_ in PyMOO:
    https://github.com/msu-coinlab/pymoo/blob/master/pymoo/vendor/hv.py

    Maximization is assumed.

    TODO: write this in C++ for faster looping.
    """

    def __init__(self, ref_point: Tensor) -> None:
        r"""Initialize hypervolume object.

        Args:
            ref_point: `m`-dim Tensor containing the reference point.

        """
        self.ref_point = ref_point

    @property
    def ref_point(self) -> Tensor:
        r"""Get reference point (for maximization).

        Returns:
            A `m`-dim tensor containing the reference point.
        """
        return -self._ref_point

    @ref_point.setter
    def ref_point(self, ref_point: Tensor) -> None:
        r"""Set the reference point for maximization

        Args:
            ref_point:  A `m`-dim tensor containing the reference point.
        """
        self._ref_point = -ref_point

    def compute(self, pareto_Y: Tensor) -> float:
        r"""Compute hypervolume.

        Args:
            pareto_Y: A `n x m`-dim tensor of pareto optimal outcomes

        Returns:
            The hypervolume.
        """
        if pareto_Y.shape[-1] != self._ref_point.shape[0]:
            raise BotorchTensorDimensionError(
                "pareto_Y must have the same number of objectives as ref_point. "
                f"Got {pareto_Y.shape[-1]}, expected {self._ref_point.shape[0]}."
            )
        elif pareto_Y.ndim != 2:
            raise BotorchTensorDimensionError(
                f"pareto_Y must have exactly two dimensions, got {pareto_Y.ndim}."
            )
        # This assumes maximization, but internally flips the sign of the pareto front
        # and the reference point and computes hypervolume for the minimization problem.
        pareto_Y = -pareto_Y
        better_than_ref = (pareto_Y <= self._ref_point).all(dim=-1)
        pareto_Y = pareto_Y[better_than_ref]
        # shift the pareto front so that reference point is all zeros
        pareto_Y = pareto_Y - self._ref_point
        self._initialize_multilist(pareto_Y)
        bounds = torch.full_like(self._ref_point, float("-inf"))
        return self._hv_recursive(
            i=self._ref_point.shape[0] - 1, n_pareto=pareto_Y.shape[0], bounds=bounds
        )

    def _hv_recursive(self, i: int, n_pareto: int, bounds: Tensor) -> float:
        r"""Recursive method for hypervolume calculation.

        This assumes minimization (internally).

        In contrast to the paper, this code assumes that the reference point
        is the origin. This enables pruning a few operations.

        Args:
            i: objective index
            n_pareto: number of pareto points
            bounds: objective bounds

        Returns:
            The hypervolume.
        """
        hvol = torch.tensor(0.0, dtype=bounds.dtype, device=bounds.device)
        sentinel = self.list.sentinel
        if n_pareto == 0:
            # base case: one dimension
            return hvol.item()
        elif i == 0:
            # base case: one dimension
            return -sentinel.next[0].data[0].item()
        elif i == 1:
            # two dimensions, end recursion
            q = sentinel.next[1]
            h = q.data[0]
            p = q.next[1]
            while p is not sentinel:
                hvol += h * (q.data[1] - p.data[1])
                if p.data[0] < h:
                    h = p.data[0]
                q = p
                p = q.next[1]
            hvol += h * q.data[1]
            return hvol.item()
        else:
            p = sentinel
            q = p.prev[i]
            while q.data is not None:
                if q.ignore < i:
                    q.ignore = 0
                q = q.prev[i]
            q = p.prev[i]
            while n_pareto > 1 and (
                q.data[i] > bounds[i] or q.prev[i].data[i] >= bounds[i]
            ):
                p = q
                self.list.remove(p, i, bounds)
                q = p.prev[i]
                n_pareto -= 1
            q_prev = q.prev[i]
            if n_pareto > 1:
                hvol = q_prev.volume[i] + q_prev.area[i] * (q.data[i] - q_prev.data[i])
            else:
                q.area[0] = 1
                q.area[1 : i + 1] = q.area[:i] * -(q.data[:i])
            q.volume[i] = hvol
            if q.ignore >= i:
                q.area[i] = q_prev.area[i]
            else:
                q.area[i] = self._hv_recursive(i - 1, n_pareto, bounds)
                if q.area[i] <= q_prev.area[i]:
                    q.ignore = i
            while p is not sentinel:
                p_data = p.data[i]
                hvol += q.area[i] * (p_data - q.data[i])
                bounds[i] = p_data
                self.list.reinsert(p, i, bounds)
                n_pareto += 1
                q = p
                p = p.next[i]
                q.volume[i] = hvol
                if q.ignore >= i:
                    q.area[i] = q.prev[i].area[i]
                else:
                    q.area[i] = self._hv_recursive(i - 1, n_pareto, bounds)
                    if q.area[i] <= q.prev[i].area[i]:
                        q.ignore = i
            hvol -= q.area[i] * q.data[i]
            return hvol.item()

    def _initialize_multilist(self, pareto_Y: Tensor) -> None:
        r"""Sets up the multilist data structure needed for calculation.

        Note: this assumes minimization.

        Args:
            pareto_Y: A `n x m`-dim tensor of pareto optimal objectives.

        """
        m = pareto_Y.shape[-1]
        nodes = [
            Node(m=m, dtype=pareto_Y.dtype, device=pareto_Y.device, data=point)
            for point in pareto_Y
        ]
        self.list = MultiList(m=m, dtype=pareto_Y.dtype, device=pareto_Y.device)
        for i in range(m):
            sort_by_dimension(nodes, i)
            self.list.extend(nodes, i)


def sort_by_dimension(nodes: list[Node], i: int) -> None:
    r"""Sorts the list of nodes in-place by the specified objective.

    Args:
        nodes: A list of Nodes
        i: The index of the objective to sort by

    """
    # build a list of tuples of (point[i], node)
    decorated = [(node.data[i], index, node) for index, node in enumerate(nodes)]
    # sort by this value
    decorated.sort()
    # write back to original list
    nodes[:] = [node for (_, _, node) in decorated]


class Node:
    r"""Node in the MultiList data structure."""

    def __init__(
        self,
        m: int,
        dtype: torch.dtype,
        device: torch.device,
        data: Tensor | None = None,
    ) -> None:
        r"""Initialize MultiList.

        Args:
            m: The number of objectives
            dtype: The dtype
            device: The device
            data: The tensor data to be stored in this Node.
        """
        self.data = data
        self.next = [None] * m
        self.prev = [None] * m
        self.ignore = 0
        self.area = torch.zeros(m, dtype=dtype, device=device)
        self.volume = torch.zeros_like(self.area)


class MultiList:
    r"""A special data structure used in hypervolume computation.

    It consists of several doubly linked lists that share common nodes.
    Every node has multiple predecessors and successors, one in every list.
    """

    def __init__(self, m: int, dtype: torch.dtype, device: torch.device) -> None:
        r"""Initialize `m` doubly linked lists.

        Args:
            m: number of doubly linked lists
            dtype: the dtype
            device: the device

        """
        self.m = m
        self.sentinel = Node(m=m, dtype=dtype, device=device)
        self.sentinel.next = [self.sentinel] * m
        self.sentinel.prev = [self.sentinel] * m

    def append(self, node: Node, index: int) -> None:
        r"""Appends a node to the end of the list at the given index.

        Args:
            node: the new node
            index: the index where the node should be appended.
        """
        last = self.sentinel.prev[index]
        node.next[index] = self.sentinel
        node.prev[index] = last
        # set the last element as the new one
        self.sentinel.prev[index] = node
        last.next[index] = node

    def extend(self, nodes: list[Node], index: int) -> None:
        r"""Extends the list at the given index with the nodes.

        Args:
            nodes: list of nodes to append at the given index.
            index: the index where the nodes should be appended.

        """
        for node in nodes:
            self.append(node=node, index=index)

    def remove(self, node: Node, index: int, bounds: Tensor) -> Node:
        r"""Removes and returns 'node' from all lists in [0, 'index'].

        Args:
            node: The node to remove
            index: The upper bound on the range of indices
            bounds: A `2 x m`-dim tensor bounds on the objectives
        """
        for i in range(index):
            predecessor = node.prev[i]
            successor = node.next[i]
            predecessor.next[i] = successor
            successor.prev[i] = predecessor
        bounds.data = torch.min(bounds, node.data)
        return node

    def reinsert(self, node: Node, index: int, bounds: Tensor) -> None:
        r"""Re-inserts the node at its original position.

        Re-inserts the node at its original position in all lists in [0, 'index']
        before it was removed. This method assumes that the next and previous
        nodes of the node that is reinserted are in the list.

        Args:
            node: The node
            index: The upper bound on the range of indices
            bounds: A `2 x m`-dim tensor bounds on the objectives

        """
        for i in range(index):
            node.prev[i].next[i] = node
            node.next[i].prev[i] = node
        bounds.data = torch.min(bounds, node.data)


class SubsetIndexCachingMixin:
    """A Mixin class that adds q-subset index computations and caching."""

    def __init__(self):
        """Initializes the class with q_out = -1 and an empty q_subset_indices dict."""
        self.q_out: int = -1
        self.q_subset_indices: BufferDict[str, Tensor] = BufferDict()

    def compute_q_subset_indices(
        self, q_out: int, device: torch.device
    ) -> BufferDict[str, Tensor]:
        r"""Returns and caches a dict of indices equal to subsets of `{1, ..., q_out}`.

        This means that consecutive calls to `self.compute_q_subset_indices` with
        the same `q_out` do not recompute the indices for all (2^q_out - 1) subsets.

        NOTE: This will use more memory than regenerating the indices
        for each i and then deleting them, but it will be faster for
        repeated evaluations (e.g. during optimization).

        Args:
            q_out: The batch size of the objectives. This is typically equal
                to the q-batch size of `X`. However, if using a set valued
                objective (e.g., MVaR) that produces `s` objective values for
                each point on the q-batch of `X`, we need to properly account
                for each objective while calculating the hypervolume contributions
                by using `q_out = q * s`.

        Returns:
            A dict that maps "q choose i" to all size-i subsets of `{1, ..., q_out}`.
        """
        if q_out != self.q_out:
            self.q_subset_indices = compute_subset_indices(q_out, device=device)
            self.q_out = q_out
        return self.q_subset_indices


def compute_subset_indices(
    q: int, device: torch.device | None = None
) -> BufferDict[str, Tensor]:
    r"""Compute all (2^q - 1) distinct subsets of {1, ..., `q`}.

    Args:
        q: An integer defininig the set {1, ..., `q`} whose subsets to compute.

    Returns:
        A dict that maps "q choose i" to all size-i subsets of {1, ..., `q_out`}.
    """
    indices = torch.arange(q, dtype=torch.long, device=device)
    return BufferDict(
        {
            f"q_choose_{i}": torch.tensor(
                list(combinations(indices, i)), dtype=torch.long, device=device
            )
            for i in range(1, q + 1)
        }
    )


class NoisyExpectedHypervolumeMixin(CachedCholeskyMCSamplerMixin):
    def __init__(
        self,
        model: Model,
        ref_point: list[float] | Tensor,
        X_baseline: Tensor,
        sampler: MCSampler | None = None,
        objective: MCMultiOutputObjective | None = None,
        constraints: list[Callable[[Tensor], Tensor]] | None = None,
        X_pending: Tensor | None = None,
        prune_baseline: bool = False,
        alpha: float = 0.0,
        cache_pending: bool = True,
        max_iep: int = 0,
        incremental_nehvi: bool = True,
        cache_root: bool = True,
        marginalize_dim: int | None = None,
    ):
        """Initialize a mixin that contains functions for the batched Pareto-frontier
        partitioning used by the noisy hypervolume-improvement-based acquisition
        functions, i.e. qNEHVI and qLogNEHVI.

        Args:
            model: A fitted model.
            ref_point: A list or tensor with `m` elements representing the reference
                point (in the outcome space) w.r.t. to which compute the hypervolume.
                This is a reference point for the objective values (i.e. after
                applying `objective` to the samples).
            X_baseline: A `r x d`-dim Tensor of `r` design points that have already
                been observed. These points are considered as potential approximate
                pareto-optimal design points.
            sampler: The sampler used to draw base samples. If not given,
                a sampler is generated using `get_sampler`. NOTE: A box decomposition is
                of the Pareto front is created for each MC sample, an operation that
                scales as `O(n^m)` and thus becomes particularly costly for `m` > 2.
            objective: The MCMultiOutputObjective under which the samples are
                evaluated. Defaults to `IdentityMCMultiOutputObjective()`.
            constraints: A list of callables, each mapping a Tensor of dimension
                `sample_shape x batch-shape x q x m` to a Tensor of dimension
                `sample_shape x batch-shape x q`, where negative values imply
                feasibility. The acqusition function will compute expected feasible
                hypervolume.
            X_pending: A `batch_shape x m x d`-dim Tensor of `m` design points that
                have points that have been submitted for function evaluation, but
                have not yet been evaluated.
            prune_baseline: If True, remove points in `X_baseline` that are
                highly unlikely to be the pareto optimal and better than the
                reference point. This can significantly improve computation time and
                is generally recommended. In order to customize pruning parameters,
                instead manually call `prune_inferior_points_multi_objective` on
                `X_baseline` before instantiating the acquisition function.
            alpha: The hyperparameter controlling the approximate non-dominated
                partitioning. The default value of 0.0 means an exact partitioning
                is used. As the number of objectives `m` increases, consider increasing
                this parameter in order to limit computational complexity.
            cache_pending: A boolean indicating whether to use cached box
                decompositions (CBD) for handling pending points. This is
                generally recommended.
            max_iep: The maximum number of pending points before the box
                decompositions will be recomputed.
            incremental_nehvi: A boolean indicating whether to compute the
                incremental NEHVI from the `i`th point where `i=1, ..., q`
                under sequential greedy optimization, or the full qNEHVI over
                `q` points.
            cache_root: A boolean indicating whether to cache the root
                decomposition over `X_baseline` and use low-rank updates.
            marginalize_dim: A batch dimension that should be marginalized. For example,
                this is useful when using a batched fully Bayesian model.
        """
        super().__init__(model=model, cache_root=cache_root, sampler=sampler)
        if len(ref_point) < 2:
            raise ValueError(
                "NoisyExpectedHypervolumeMixin supports m>=2 outcomes "
                f"but ref_point has length {len(ref_point)}, which is smaller than 2."
            )
        tkwargs = {"dtype": X_baseline.dtype, "device": X_baseline.device}
        ref_point = torch.as_tensor(ref_point, **tkwargs)
        self.register_buffer("ref_point", ref_point)

        if X_baseline.ndim > 2:
            raise UnsupportedError(
                f"NoisyExpectedHypervolumeMixin does not support batched "
                f"X_baseline. Expected 2 dims, got {X_baseline.ndim}."
            )
        if prune_baseline:
            X_baseline = prune_inferior_points_multi_objective(
                model=model,
                X=X_baseline,
                objective=objective,
                constraints=constraints,
                ref_point=ref_point,
                marginalize_dim=marginalize_dim,
            )

        self.alpha = alpha
        self.q_in = -1
        self.q_out = -1

        self.partitioning = None
        # set partitioning class and args
        self.p_kwargs = {}
        if self.alpha > 0:
            self.p_kwargs["alpha"] = self.alpha
            self.p_class = NondominatedPartitioning
        else:
            self.p_class = FastNondominatedPartitioning
        self.register_buffer("_X_baseline", X_baseline)
        self.register_buffer("_X_baseline_and_pending", X_baseline)
        self.register_buffer(
            "cache_pending",
            torch.tensor(cache_pending, dtype=bool),
        )
        self.register_buffer(
            "_prev_nehvi",
            torch.tensor(0.0, **tkwargs),
        )
        self.register_buffer(
            "_max_iep",
            torch.tensor(max_iep, dtype=torch.long),
        )
        self.register_buffer(
            "incremental_nehvi",
            torch.tensor(incremental_nehvi, dtype=torch.bool),
        )
        # Base sampler is initialized in _set_cell_bounds.
        self.base_sampler = None

        # is this called twice, once here, once in MultiObjectiveMCAcquisitionFunction?
        if X_pending is not None:
            # This will call self._set_cell_bounds if the number of pending
            # points is greater than self._max_iep.
            self.set_X_pending(X_pending)
        # In the case that X_pending is not None, but there are fewer than
        # max_iep pending points, the box decompositions are not performed in
        # set_X_pending. Therefore, we need to perform a box decomposition over
        # f(X_baseline) here.
        if X_pending is None or X_pending.shape[-2] <= self._max_iep:
            self._set_cell_bounds(num_new_points=X_baseline.shape[0])

        # Set q_in=-1 to so that self.sampler is updated at the next forward call.
        self.q_in = -1

    @property
    def X_baseline(self) -> Tensor:
        r"""Return X_baseline augmented with pending points cached using CBD."""
        return self._X_baseline_and_pending

    def _compute_initial_hvs(self, obj: Tensor, feas: Tensor | None = None) -> None:
        r"""Compute hypervolume dominated by f(X_baseline) under each sample.

        Args:
            obj: A `sample_shape x batch_shape x n x m`-dim tensor of samples
                of objectives.
            feas: `sample_shape x batch_shape x n`-dim tensor of samples
                of feasibility indicators.
        """
        initial_hvs = []
        for i, sample in enumerate(obj):
            if self.constraints is not None:
                sample = sample[feas[i]]
            dominated_partitioning = DominatedPartitioning(
                ref_point=self.ref_point,
                Y=sample,
            )
            hv = dominated_partitioning.compute_hypervolume()
            initial_hvs.append(hv)
        self.register_buffer(
            "_initial_hvs",
            torch.tensor(initial_hvs, dtype=obj.dtype, device=obj.device).view(
                self._batch_sample_shape, *obj.shape[-2:]
            ),
        )

    def _set_cell_bounds(self, num_new_points: int) -> None:
        r"""Compute the box decomposition under each posterior sample.

        Args:
            num_new_points: The number of new points (beyond the points
                in X_baseline) that were used in the previous box decomposition.
                In the first box decomposition, this should be the number of points
                in X_baseline.
        """
        if self.X_baseline.shape[0] > 0:
            with torch.no_grad():
                posterior = self.model.posterior(self.X_baseline)
            # Reset sampler, accounting for possible one-to-many transform.
            self.q_in = -1
            if self.base_sampler is None:
                # Initialize the base sampler if needed.
                samples = self.get_posterior_samples(posterior)
                self.base_sampler = deepcopy(self.sampler)
            else:
                samples = self.base_sampler(posterior)
            n_w = posterior._extended_shape()[-2] // self.X_baseline.shape[-2]
            self._set_sampler(q_in=num_new_points * n_w, posterior=posterior)
            # cache posterior
            if self._cache_root:
                # Note that this implicitly uses LinearOperator's caching to check if
                # the proper root decomposition has already been cached to
                # `posterior.mvn.lazy_covariance_matrix`, which it may have been in
                # the call to `self.base_sampler`, and computes it if not found
                self._baseline_L = self._compute_root_decomposition(posterior=posterior)
            obj = self.objective(samples, X=self.X_baseline)

        else:
            sample_shape = (
                self.sampler.sample_shape
                if self.sampler is not None
                else self._default_sample_shape
            )
            obj = torch.empty(
                *sample_shape,
                0,
                self.ref_point.shape[-1],
                dtype=self.ref_point.dtype,
                device=self.ref_point.device,
            )

        # compute feasibility indicator if there are constraints
        if self.constraints is None or self.X_baseline.shape[0] == 0:
            feas = None
        else:
            feas = compute_feasibility_indicator(
                constraints=self.constraints, samples=samples
            )

        self._batch_sample_shape = obj.shape[:-2]
        # collapse batch dimensions
        # use numel() rather than view(-1) to handle case of no baseline points
        new_batch_shape = self._batch_sample_shape.numel()
        obj = obj.view(new_batch_shape, *obj.shape[-2:])
        if feas is not None:
            feas = feas.view(new_batch_shape, *feas.shape[-1:])

        if self.partitioning is None and not self.incremental_nehvi:
            self._compute_initial_hvs(obj=obj, feas=feas)

        if self.ref_point.shape[-1] > 2:
            # the partitioning algorithms run faster on the CPU
            # due to advanced indexing
            ref_point_cpu = self.ref_point.cpu()
            obj_cpu = obj.cpu()
            if feas is not None:
                feas_cpu = feas.cpu()
                obj_cpu = [obj_cpu[i][feas_cpu[i]] for i in range(obj.shape[0])]
            partitionings = []
            for sample in obj_cpu:
                partitioning = self.p_class(
                    ref_point=ref_point_cpu, Y=sample, **self.p_kwargs
                )
                partitionings.append(partitioning)
            self.partitioning = BoxDecompositionList(*partitionings)
        else:
            # use batched partitioning
            obj = _pad_batch_pareto_frontier(
                Y=obj,
                ref_point=self.ref_point.unsqueeze(0).expand(
                    obj.shape[0], self.ref_point.shape[-1]
                ),
                feasibility_mask=feas,
            )
            self.partitioning = self.p_class(
                ref_point=self.ref_point, Y=obj, **self.p_kwargs
            )
        cell_bounds = self.partitioning.get_hypercell_bounds().to(self.ref_point)
        cell_bounds = cell_bounds.view(
            2, *self._batch_sample_shape, *cell_bounds.shape[-2:]
        )  # 2 x batch_shape x sample_shape x num_cells x m
        self.register_buffer("cell_lower_bounds", cell_bounds[0])
        self.register_buffer("cell_upper_bounds", cell_bounds[1])

    def set_X_pending(self, X_pending: Tensor | None = None) -> None:
        r"""Informs the acquisition function about pending design points.

        Args:
            X_pending: `n x d` Tensor with `n` `d`-dim design points that have
                been submitted for evaluation but have not yet been evaluated.
        """
        if X_pending is None:
            self.X_pending = None
        else:
            if X_pending.requires_grad:
                warnings.warn(
                    "Pending points require a gradient but the acquisition function"
                    " will not provide a gradient to these points.",
                    BotorchWarning,
                    stacklevel=2,
                )
            X_pending = X_pending.detach().clone()
            if self.cache_pending:
                X_baseline = torch.cat([self._X_baseline, X_pending], dim=-2)
                # Number of new points is the total number of points minus
                # (the number of previously cached pending points plus the
                # of number of baseline points).
                num_new_points = X_baseline.shape[0] - self.X_baseline.shape[0]
                if num_new_points > 0:
                    if num_new_points > self._max_iep:
                        # Set the new baseline points to include pending points.
                        self.register_buffer("_X_baseline_and_pending", X_baseline)
                        # Recompute box decompositions.
                        self._set_cell_bounds(num_new_points=num_new_points)
                        if not self.incremental_nehvi:
                            self._prev_nehvi = (
                                (self._hypervolumes - self._initial_hvs)
                                .clamp_min(0.0)
                                .mean()
                            )
                        # Set to None so that pending points are not concatenated in
                        # forward.
                        self.X_pending = None
                        # Set q_in=-1 to so that self.sampler is updated at the next
                        # forward call.
                        self.q_in = -1
                    else:
                        self.X_pending = X_pending[-num_new_points:]
            else:
                self.X_pending = X_pending

    @property
    def _hypervolumes(self) -> Tensor:
        r"""Compute hypervolume over X_baseline under each posterior sample.

        Returns:
            A `sample_shape`-dim tensor of hypervolumes.
        """
        return (
            self.partitioning.compute_hypervolume()
            .to(self.ref_point)  # for m > 2, the partitioning is on the CPU
            .view(self._batch_sample_shape)
        )


def get_hypervolume_maximizing_subset(
    n: int, Y: Tensor, ref_point: Tensor
) -> tuple[Tensor, Tensor]:
    """Find an approximately hypervolume-maximizing subset of size `n`.

    This greedily selects points from Y to maximize the hypervolume of
    the subset sequentially. This has bounded error since hypervolume is
    submodular.

    Args:
        n: The size of the subset to return.
        Y: A `n' x m`-dim tensor of outcomes.
        ref_point: A `m`-dim tensor containing the reference point.

    Returns:
        A two-element tuple containing
            - A `n x m`-dim tensor of outcomes.
            - A `n`-dim tensor of indices of the outcomes in the original set.
    """
    if Y.ndim != 2:
        raise NotImplementedError(
            "Only two dimensions are supported (no additional) batch dims."
        )
    elif Y.shape[0] < n:
        raise ValueError(
            f"Y has fewer points ({Y.shape[0]}) than the requested subset size ({n})."
        )
    Y_subset = torch.zeros(0, Y.shape[1], dtype=Y.dtype, device=Y.device)
    selected_indices = []
    remaining_idcs = set(range(Y.shape[0]))
    best_hv = 0.0
    for _ in range(n):
        # Add each point and compute the hypervolume
        best_idx = None
        for i in remaining_idcs:
            partitioning = DominatedPartitioning(
                ref_point=ref_point, Y=torch.cat((Y_subset, Y[i : i + 1]), dim=0)
            )
            hv = partitioning.compute_hypervolume().item()
            if hv > best_hv:
                best_idx = i
                best_hv = hv
        if best_idx is None:
            # no arm improved HV, so select a random arm. This will only happen if Y is
            # not a Pareto frontier, where all points are better than the reference
            # point
            best_idx = random.choice(list(remaining_idcs))
        remaining_idcs.remove(best_idx)
        selected_indices.append(best_idx)
        Y_subset = torch.cat((Y_subset, Y[best_idx : best_idx + 1]), dim=0)
    return Y_subset, torch.tensor(selected_indices, dtype=torch.long, device=Y.device)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Posterior module to be used with GPyTorch models.
"""

from __future__ import annotations

from contextlib import ExitStack
from typing import TYPE_CHECKING

import torch
from botorch.exceptions.errors import BotorchTensorDimensionError
from botorch.posteriors.base_samples import _reshape_base_samples_non_interleaved
from botorch.posteriors.torch import TorchPosterior
from gpytorch.distributions import MultitaskMultivariateNormal, MultivariateNormal
from linear_operator import settings as linop_settings
from linear_operator.operators import (
    BlockDiagLinearOperator,
    DenseLinearOperator,
    LinearOperator,
    SumLinearOperator,
)
from torch import Tensor
from torch.distributions import Normal

if TYPE_CHECKING:
    from botorch.posteriors.posterior_list import PosteriorList  # pragma: no cover


class GPyTorchPosterior(TorchPosterior):
    r"""A posterior based on GPyTorch's multi-variate Normal distributions."""

    distribution: MultivariateNormal

    def __init__(self, distribution: MultivariateNormal) -> None:
        r"""A posterior based on GPyTorch's multi-variate Normal distributions.

        Args:
            distribution: A GPyTorch MultivariateNormal (single-output case) or
                MultitaskMultivariateNormal (multi-output case).
        """
        super().__init__(distribution=distribution)
        self._is_mt: bool = isinstance(distribution, MultitaskMultivariateNormal)

    @property
    def mvn(self) -> MultivariateNormal:
        r"""Expose the distribution as a backwards-compatible attribute."""
        return self.distribution

    @property
    def base_sample_shape(self) -> torch.Size:
        r"""The shape of a base sample used for constructing posterior samples."""
        return self.distribution.batch_shape + self.distribution.base_sample_shape

    @property
    def batch_range(self) -> tuple[int, int]:
        r"""The t-batch range.

        This is used in samplers to identify the t-batch component of the
        `base_sample_shape`. The base samples are expanded over the t-batches to
        provide consistency in the acquisition values, i.e., to ensure that a
        candidate produces same value regardless of its position on the t-batch.
        """
        if self._is_mt:
            return (0, -2)
        else:
            return (0, -1)

    def _extended_shape(
        self,
        sample_shape: torch.Size = torch.Size(),  # noqa: B008
    ) -> torch.Size:
        r"""Returns the shape of the samples produced by the posterior with
        the given `sample_shape`.
        """
        base_shape = self.distribution.batch_shape + self.distribution.event_shape
        if not self._is_mt:
            base_shape += torch.Size([1])
        return sample_shape + base_shape

    def rsample_from_base_samples(
        self,
        sample_shape: torch.Size,
        base_samples: Tensor,
    ) -> Tensor:
        r"""Sample from the posterior (with gradients) using base samples.

        This is intended to be used with a sampler that produces the corresponding base
        samples, and enables acquisition optimization via Sample Average Approximation.

        Args:
            sample_shape: A `torch.Size` object specifying the sample shape. To
                draw `n` samples, set to `torch.Size([n])`. To draw `b` batches
                of `n` samples each, set to `torch.Size([b, n])`.
            base_samples: A Tensor of `N(0, I)` base samples of shape
                `sample_shape x base_sample_shape`, typically obtained from
                a `Sampler`. This is used for deterministic optimization.

        Returns:
            Samples from the posterior, a tensor of shape
            `self._extended_shape(sample_shape=sample_shape)`.
        """
        if base_samples.shape[: len(sample_shape)] != sample_shape:
            raise RuntimeError(
                "`sample_shape` disagrees with shape of `base_samples`. "
                f"Got {sample_shape=} and {base_samples.shape=}."
            )
        if self._is_mt:
            base_samples = _reshape_base_samples_non_interleaved(
                mvn=self.distribution,
                base_samples=base_samples,
                sample_shape=sample_shape,
            )
        with ExitStack() as es:
            if linop_settings._fast_covar_root_decomposition.is_default():
                es.enter_context(linop_settings._fast_covar_root_decomposition(False))
            samples = self.distribution.rsample(
                sample_shape=sample_shape, base_samples=base_samples
            )
        if not self._is_mt:
            samples = samples.unsqueeze(-1)
        return samples

    def rsample(self, sample_shape: torch.Size | None = None) -> Tensor:
        r"""Sample from the posterior (with gradients).

        Args:
            sample_shape: A `torch.Size` object specifying the sample shape. To
                draw `n` samples, set to `torch.Size([n])`. To draw `b` batches
                of `n` samples each, set to `torch.Size([b, n])`.

        Returns:
            Samples from the posterior, a tensor of shape
            `self._extended_shape(sample_shape=sample_shape)`.
        """
        if sample_shape is None:
            sample_shape = torch.Size([1])
        with ExitStack() as es:
            if linop_settings._fast_covar_root_decomposition.is_default():
                es.enter_context(linop_settings._fast_covar_root_decomposition(False))
            samples = self.distribution.rsample(sample_shape=sample_shape)
        # make sure there always is an output dimension
        if not self._is_mt:
            samples = samples.unsqueeze(-1)
        return samples

    @property
    def mean(self) -> Tensor:
        r"""The posterior mean."""
        mean = self.distribution.mean
        if not self._is_mt:
            mean = mean.unsqueeze(-1)
        return mean

    @property
    def variance(self) -> Tensor:
        r"""The posterior variance."""
        variance = self.distribution.variance
        if not self._is_mt:
            variance = variance.unsqueeze(-1)
        return variance

    def quantile(self, value: Tensor) -> Tensor:
        r"""Compute the quantiles of the marginal distributions."""
        if value.numel() > 1:
            return torch.stack([self.quantile(v) for v in value], dim=0)
        marginal = Normal(loc=self.mean, scale=self.variance.sqrt())
        return marginal.icdf(value)

    def density(self, value: Tensor) -> Tensor:
        r"""The probability density of the marginal distributions."""
        if value.numel() > 1:
            return torch.stack([self.density(v) for v in value], dim=0)
        marginal = Normal(loc=self.mean, scale=self.variance.sqrt())
        return marginal.log_prob(value).exp()


def _validate_scalarize_inputs(weights: Tensor, m: int) -> None:
    if weights.ndim > 1:
        raise BotorchTensorDimensionError("`weights` must be one-dimensional")
    if m != weights.size(0):
        raise RuntimeError(
            f"Output shape not equal to that of weights. Output shape is {m} and "
            f"weights are {weights.shape}"
        )


def scalarize_posterior_gpytorch(
    posterior: GPyTorchPosterior,
    weights: Tensor,
    offset: float = 0.0,
) -> tuple[Tensor, Tensor | LinearOperator]:
    r"""Helper function for `scalarize_posterior`, producing a mean and
    variance.

    This mean and variance are consumed by `scalarize_posterior` to produce
    a `GPyTorchPosterior`.

    Args:
        posterior: The posterior over `m` outcomes to be scalarized.
            Supports `t`-batching.
        weights: A tensor of weights of size `m`.
        offset: The offset of the affine transformation.

    Returns:
        The transformed (single-output) posterior. If the input posterior has
            mean `mu` and covariance matrix `Sigma`, this posterior has mean
            `weights^T * mu` and variance `weights^T Sigma w`.

    Example:
        Example for a model with two outcomes:

        >>> X = torch.rand(1, 2)
        >>> posterior = model.posterior(X)
        >>> weights = torch.tensor([0.5, 0.25])
        >>> mean, cov = scalarize_posterior_gpytorch(posterior, weights=weights)
        >>> mvn = MultivariateNormal(mean, cov)
        >>> new_posterior = GPyTorchPosterior
    """
    mean = posterior.mean
    q, m = mean.shape[-2:]
    _validate_scalarize_inputs(weights=weights, m=m)
    batch_shape = mean.shape[:-2]
    mvn = posterior.distribution
    cov = mvn.lazy_covariance_matrix if mvn.islazy else mvn.covariance_matrix

    if m == 1:  # just scaling, no scalarization necessary
        new_mean = offset + (weights[0] * mean).view(*batch_shape, q)
        new_cov = weights[0] ** 2 * cov
        return new_mean, new_cov

    new_mean = offset + (mean @ weights).view(*batch_shape, q)

    if q == 1:
        new_cov = weights.unsqueeze(-2) @ (cov @ weights.unsqueeze(-1))
    else:
        # we need to handle potentially different representations of the multi-task mvn
        if mvn._interleaved:
            w_cov = weights.repeat(q).unsqueeze(0)
            sum_shape = batch_shape + torch.Size([q, m, q, m])
            sum_dims = (-1, -2)
        else:
            # special-case the independent setting
            if isinstance(cov, BlockDiagLinearOperator):
                new_cov = SumLinearOperator(
                    *[
                        cov.base_linear_op[..., i, :, :] * weights[i].pow(2)
                        for i in range(cov.base_linear_op.size(-3))
                    ]
                )
                return new_mean, new_cov

            w_cov = torch.repeat_interleave(weights, q).unsqueeze(0)
            sum_shape = batch_shape + torch.Size([m, q, m, q])
            sum_dims = (-2, -3)

        cov_scaled = w_cov * cov * w_cov.transpose(-1, -2)
        # TODO: Do not instantiate full covariance for LinearOperators
        # (ideally we simplify this in GPyTorch:
        # https://github.com/cornellius-gp/gpytorch/issues/1055)
        if isinstance(cov_scaled, LinearOperator):
            cov_scaled = cov_scaled.to_dense()
        new_cov = cov_scaled.view(sum_shape).sum(dim=sum_dims[0]).sum(dim=sum_dims[1])
        new_cov = DenseLinearOperator(new_cov)

    return new_mean, new_cov


def scalarize_posterior(
    posterior: GPyTorchPosterior | PosteriorList,
    weights: Tensor,
    offset: float = 0.0,
) -> GPyTorchPosterior:
    r"""Affine transformation of a multi-output posterior.

    Args:
        posterior: The posterior over `m` outcomes to be scalarized.
            Supports `t`-batching. Can be either a `GPyTorchPosterior`,
            or a `PosteriorList` that contains GPyTorchPosteriors all with q=1.
        weights: A tensor of weights of size `m`.
        offset: The offset of the affine transformation.

    Returns:
        The transformed (single-output) posterior. If the input posterior has
            mean `mu` and covariance matrix `Sigma`, this posterior has mean
            `weights^T * mu` and variance `weights^T Sigma w`.

    Example:
        Example for a model with two outcomes:

        >>> X = torch.rand(1, 2)
        >>> posterior = model.posterior(X)
        >>> weights = torch.tensor([0.5, 0.25])
        >>> new_posterior = scalarize_posterior(posterior, weights=weights)
    """
    # GPyTorchPosterior case
    if hasattr(posterior, "distribution"):
        mean, cov = scalarize_posterior_gpytorch(posterior, weights, offset)
        mvn = MultivariateNormal(mean, cov)
        return GPyTorchPosterior(mvn)

    # PosteriorList case
    if not hasattr(posterior, "posteriors"):
        raise NotImplementedError(
            "scalarize_posterior only works with a posterior that has an attribute "
            "`distribution`, such as a GPyTorchPosterior, or a posterior that contains "
            "sub-posteriors in an attribute `posteriors`, as in a PosteriorList."
        )

    mean = posterior.mean
    q, m = mean.shape[-2:]

    _validate_scalarize_inputs(weights, m)
    batch_shape = mean.shape[:-2]

    if q != 1:
        raise NotImplementedError(
            "scalarize_posterior only works with a PosteriorList if each sub-posterior "
            "has q=1."
        )

    means = [post.mean for post in posterior.posteriors]
    if {mean.shape[-1] for mean in means} != {1}:
        raise NotImplementedError(
            "scalarize_posterior only works with a PosteriorList if each sub-posterior "
            "has one outcome."
        )

    new_mean = offset + (mean @ weights).view(*batch_shape, q)
    new_cov = (posterior.variance @ (weights**2))[:, None]
    mvn = MultivariateNormal(new_mean, new_cov)
    return GPyTorchPosterior(mvn)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import warnings

from collections.abc import Callable

from typing import Any

import torch
from botorch.acquisition import AcquisitionFunction

from botorch.generation.gen import TGenCandidates
from botorch.optim.homotopy import Homotopy
from botorch.optim.initializers import TGenInitialConditions
from botorch.optim.optimize import optimize_acqf, optimize_acqf_mixed
from torch import Tensor


def prune_candidates(
    candidates: Tensor, acq_values: Tensor, prune_tolerance: float
) -> Tensor:
    r"""Prune candidates based on their distance to other candidates.

    Args:
        candidates: An `n x d` tensor of candidates.
        acq_values: An `n` tensor of candidate values.
        prune_tolerance: The minimum distance to prune candidates.

    Returns:
        An `m x d` tensor of pruned candidates.
    """
    if candidates.ndim != 2:
        raise ValueError("`candidates` must be of size `n x d`.")
    if acq_values.ndim != 1 or len(acq_values) != candidates.shape[0]:
        raise ValueError("`acq_values` must be of size `n`.")
    if prune_tolerance < 0:
        raise ValueError("`prune_tolerance` must be >= 0.")
    sorted_inds = acq_values.argsort(descending=True)
    candidates = candidates[sorted_inds]

    candidates_new = candidates[:1, :]
    for i in range(1, candidates.shape[0]):
        if (
            torch.cdist(candidates[i : i + 1, :], candidates_new).min()
            > prune_tolerance
        ):
            candidates_new = torch.cat(
                [candidates_new, candidates[i : i + 1, :]], dim=-2
            )
    return candidates_new


def optimize_acqf_homotopy(
    acq_function: AcquisitionFunction,
    bounds: Tensor,
    q: int,
    num_restarts: int,
    homotopy: Homotopy,
    prune_tolerance: float = 1e-4,
    raw_samples: int | None = None,
    options: dict[str, bool | float | int | str] | None = None,
    final_options: dict[str, bool | float | int | str] | None = None,
    inequality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
    equality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
    nonlinear_inequality_constraints: list[tuple[Callable, bool]] | None = None,
    fixed_features: dict[int, float] | None = None,
    fixed_features_list: list[dict[int, float]] | None = None,
    post_processing_func: Callable[[Tensor], Tensor] | None = None,
    batch_initial_conditions: Tensor | None = None,
    gen_candidates: TGenCandidates | None = None,
    *,
    ic_generator: TGenInitialConditions | None = None,
    timeout_sec: float | None = None,
    retry_on_optimization_warning: bool = True,
    **ic_gen_kwargs: Any,
) -> tuple[Tensor, Tensor]:
    r"""Generate a set of candidates via multi-start optimization.

    Args:
        acq_function: An AcquisitionFunction.
        bounds: A `2 x d` tensor of lower and upper bounds for each column of `X`
            (if inequality_constraints is provided, these bounds can be -inf and
            +inf, respectively).
        q: The number of candidates.
        homotopy: Homotopy object that will make the necessary modifications to the
            problem when calling `step()`.
        prune_tolerance: The minimum distance to prune candidates.
        num_restarts: The number of starting points for multistart acquisition
            function optimization.
        raw_samples: The number of samples for initialization. This is required
            if `batch_initial_conditions` is not specified.
        options: Options for candidate generation in the initial step of the homotopy.
        final_options: Options for candidate generation in the final step of
            the homotopy.
        inequality_constraints: A list of tuples (indices, coefficients, rhs),
            with each tuple encoding an inequality constraint of the form
            `\sum_i (X[indices[i]] * coefficients[i]) >= rhs`. `indices` and
            `coefficients` should be torch tensors. See the docstring of
            `make_scipy_linear_constraints` for an example. When q=1, or when
            applying the same constraint to each candidate in the batch
            (intra-point constraint), `indices` should be a 1-d tensor.
            For inter-point constraints, in which the constraint is applied to the
            whole batch of candidates, `indices` must be a 2-d tensor, where
            in each row `indices[i] =(k_i, l_i)` the first index `k_i` corresponds
            to the `k_i`-th element of the `q`-batch and the second index `l_i`
            corresponds to the `l_i`-th feature of that element.
        equality_constraints: A list of tuples (indices, coefficients, rhs),
            with each tuple encoding an equality constraint of the form
            `\sum_i (X[indices[i]] * coefficients[i]) = rhs`. See the docstring of
            `make_scipy_linear_constraints` for an example.
        nonlinear_inequality_constraints: A list of tuples representing the nonlinear
            inequality constraints. The first element in the tuple is a callable
            representing a constraint of the form `callable(x) >= 0`. In case of an
            intra-point constraint, `callable()`takes in an one-dimensional tensor of
            shape `d` and returns a scalar. In case of an inter-point constraint,
            `callable()` takes a two dimensional tensor of shape `q x d` and again
            returns a scalar. The second element is a boolean, indicating if it is an
            intra-point or inter-point constraint (`True` for intra-point. `False` for
            inter-point). For more information on intra-point vs inter-point
            constraints, see the docstring of the `inequality_constraints` argument to
            `optimize_acqf()`. The constraints will later be passed to the scipy
            solver. You need to pass in `batch_initial_conditions` in this case.
            Using non-linear inequality constraints also requires that `batch_limit`
            is set to 1, which will be done automatically if not specified in
            `options`.
        fixed_features: A map `{feature_index: value}` for features that
            should be fixed to a particular value during generation.
        fixed_features_list: A list of maps `{feature_index: value}`. The i-th
            item represents the fixed_feature for the i-th optimization. If
            `fixed_features_list` is provided, `optimize_acqf_mixed` is invoked.
            All indices (`feature_index`) should be non-negative.
        post_processing_func: A function that post-processes an optimization
            result appropriately (i.e., according to `round-trip`
            transformations).
        batch_initial_conditions: A tensor to specify the initial conditions. Set
            this if you do not want to use default initialization strategy.
        gen_candidates: A callable for generating candidates (and their associated
            acquisition values) given a tensor of initial conditions and an
            acquisition function. Other common inputs include lower and upper bounds
            and a dictionary of options, but refer to the documentation of specific
            generation functions (e.g gen_candidates_scipy and gen_candidates_torch)
            for method-specific inputs. Default: `gen_candidates_scipy`
        ic_generator: Function for generating initial conditions. Not needed when
            `batch_initial_conditions` are provided. Defaults to
            `gen_one_shot_kg_initial_conditions` for `qKnowledgeGradient` acquisition
            functions and `gen_batch_initial_conditions` otherwise. Must be specified
            for nonlinear inequality constraints.
        timeout_sec: Max amount of time optimization can run for.
        retry_on_optimization_warning: Whether to retry candidate generation with a new
            set of initial conditions when it fails with an `OptimizationWarning`.
        ic_gen_kwargs: Additional keyword arguments passed to function specified by
            `ic_generator`
    """
    if fixed_features and fixed_features_list:
        raise ValueError(
            "Either `fixed_feature` or `fixed_features_list` can be provided, not both."
        )

    if fixed_features:
        message = (
            "The `fixed_features` argument is deprecated, "
            "use `fixed_features_list` instead."
        )
        warnings.warn(
            message,
            DeprecationWarning,
            stacklevel=2,
        )

    shared_optimize_acqf_kwargs = {
        "num_restarts": num_restarts,
        "inequality_constraints": inequality_constraints,
        "equality_constraints": equality_constraints,
        "nonlinear_inequality_constraints": nonlinear_inequality_constraints,
        "return_best_only": False,  # False to make n_restarts persist through homotopy.
        "gen_candidates": gen_candidates,
        "ic_generator": ic_generator,
        "timeout_sec": timeout_sec,
        "retry_on_optimization_warning": retry_on_optimization_warning,
        **ic_gen_kwargs,
    }

    if fixed_features_list and len(fixed_features_list) > 1:
        optimization_fn = optimize_acqf_mixed
        fixed_features_kwargs = {"fixed_features_list": fixed_features_list}
    else:
        optimization_fn = optimize_acqf
        fixed_features_kwargs = {
            "fixed_features": fixed_features_list[0]
            if fixed_features_list
            else fixed_features
        }

    candidate_list, acq_value_list = [], []
    if q > 1:
        base_X_pending = acq_function.X_pending

    for _ in range(q):
        candidates = batch_initial_conditions
        q_raw_samples = raw_samples
        homotopy.restart()

        while not homotopy.should_stop:
            candidates, acq_values = optimization_fn(
                acq_function=acq_function,
                bounds=bounds,
                q=1,
                options=options,
                batch_initial_conditions=candidates,
                raw_samples=q_raw_samples,
                **fixed_features_kwargs,
                **shared_optimize_acqf_kwargs,
            )

            homotopy.step()

            # Set raw_samples to None such that pruned restarts are not repopulated
            # at each step in the homotopy.
            q_raw_samples = None

            # Prune candidates
            candidates = prune_candidates(
                candidates=candidates.squeeze(1),
                acq_values=acq_values,
                prune_tolerance=prune_tolerance,
            ).unsqueeze(1)

        # Optimize one more time with the final options
        candidates, acq_values = optimization_fn(
            acq_function=acq_function,
            bounds=bounds,
            q=1,
            options=final_options,
            raw_samples=q_raw_samples,
            batch_initial_conditions=candidates,
            **fixed_features_kwargs,
            **shared_optimize_acqf_kwargs,
        )

        # Post-process the candidates and grab the best candidate
        if post_processing_func is not None:
            candidates = post_processing_func(candidates)
            acq_values = acq_function(candidates)

        best = torch.argmax(acq_values.view(-1), dim=0)
        candidate, acq_value = candidates[best], acq_values[best]

        # Keep the new candidate and update the pending points
        candidate_list.append(candidate)
        acq_value_list.append(acq_value)
        selected_candidates = torch.cat(candidate_list, dim=-2)

        if q > 1:
            acq_function.set_X_pending(
                torch.cat([base_X_pending, selected_candidates], dim=-2)
                if base_X_pending is not None
                else selected_candidates
            )

    if q > 1:  # Reset acq_function to previous X_pending state
        acq_function.set_X_pending(base_X_pending)

    homotopy.reset()  # Reset the homotopy parameters

    return selected_candidates, torch.stack(acq_value_list)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from collections.abc import Callable

from inspect import getsource, getsourcefile
from typing import Any

from multipledispatch.dispatcher import (
    Dispatcher as MDDispatcher,
    MDNotImplementedError,  # trivial subclass of NotImplementedError
    str_signature,
)


def type_bypassing_encoder(arg: Any) -> type:
    # Allow type variables to be passed as pre-encoded arguments
    return arg if isinstance(arg, type) else type(arg)


class Dispatcher(MDDispatcher):
    r"""Clearing house for multiple dispatch functionality. This class extends
    `<multipledispatch.Dispatcher>` by: (i) generalizing the argument encoding
    convention during method lookup, (ii) implementing `__getitem__` as a dedicated
    method lookup function.
    """

    def __init__(
        self,
        name: str,
        doc: str | None = None,
        encoder: Callable[Any, type] = type,
    ) -> None:
        """
        Args:
            name: A string identifier for the `Dispatcher` instance.
            doc: A docstring for the multiply dispatched method(s).
            encoder: A callable that individually transforms the arguments passed
                at runtime in order to construct the key used for method lookup as
                `tuple(map(encoder, args))`. Defaults to `type`.
        """
        super().__init__(name=name, doc=doc)
        self._encoder = encoder

    def __getitem__(
        self,
        args: Any | None = None,
        types: tuple[type] | None = None,
    ) -> Callable:
        r"""Method lookup.

        Args:
            args: A set of arguments that act as identifiers for a stored method.
            types: A tuple of types that encodes `args`.

        Returns:
            A callable corresponding to the given `args` or `types`.
        """
        if types is None:
            if args is None:
                raise RuntimeError("One of `args` or `types` must be provided.")
            types = self.encode_args(args)
        elif args is not None:
            raise RuntimeError("Only one of `args` or `types` may be provided.")

        try:
            func = self._cache[types]
        except KeyError:
            func = self.dispatch(*types)
            if not func:
                msg = f"{self.name}: <{', '.join(cls.__name__ for cls in types)}"
                raise NotImplementedError(f"Could not find signature for {msg}")
            self._cache[types] = func
        return func

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        r"""Multiply dispatches a call to a collection of methods.

        Args:
            args: A set of arguments that act as identifiers for a stored method.
            kwargs: Optional keyword arguments passed to the retrieved method.

        Returns:
            The result of evaluating `func(*args, **kwargs)`, where `func` is
            the function obtained via method lookup.
        """
        types = self.encode_args(args)
        func = self.__getitem__(types=types)
        try:
            return func(*args, **kwargs)
        except MDNotImplementedError:
            # Traverses registered methods in order, yields whenever a match is found
            funcs = self.dispatch_iter(*types)
            next(funcs)  # burn first, same as self.__getitem__(types=types)
            for func in funcs:
                try:
                    return func(*args, **kwargs)
                except MDNotImplementedError:
                    pass

            raise NotImplementedError(
                f"Matching functions for {self.name:s}: {str_signature(types):s} "
                "found, but none completed successfully"
            )

    def dispatch(self, *types: type) -> Callable:
        r"""Method lookup strategy. Checks for an exact match before traversing
        the set of registered methods according to the current ordering.

        Args:
            types: A tuple of types that gets compared with the signatures
                of registered methods to determine compatibility.

        Returns:
            The first method encountered with a matching signature.
        """
        if types in self.funcs:
            return self.funcs[types]
        try:
            return next(self.dispatch_iter(*types))
        except StopIteration:
            return None

    def encode_args(self, args: Any) -> tuple[type]:
        r"""Converts arguments into a tuple of types used during method lookup."""
        return tuple(map(self.encoder, args if isinstance(args, tuple) else (args,)))

    def _help(self, *args: Any) -> str:
        r"""Returns the retrieved method's docstring."""
        return self.dispatch(*self.encode_args(args)).__doc__

    def help(self, *args: Any, **kwargs: Any) -> None:
        r"""Prints the retrieved method's docstring."""
        print(self._help(*args))

    def _source(self, *args: Any) -> str:
        r"""Returns the retrieved method's source types as a string."""
        func = self.dispatch(*self.encode_args(args))
        if not func:
            raise TypeError("No function found")
        return f"File: {getsourcefile(func)}\n\n{getsource(func)}"

    def source(self, *args, **kwargs) -> None:
        r"""Prints the retrieved method's source types."""
        print(self._source(*args))

    @property
    def encoder(self) -> Callable[Any, type]:
        return self._encoder



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
The following code is from the repository vbll (https://github.com/VectorInstitute/vbll)
which is under the MIT license.
Paper: "Variational Bayesian Last Layers" by Harrison et al., ICLR 2024
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Callable

import numpy as np

import torch
import torch.nn as nn

from botorch.logging import logger
from torch import Tensor


def tp(M):
    return M.transpose(-1, -2)


class Normal(torch.distributions.Normal):
    def __init__(self, loc: Tensor, chol: Tensor):
        """Normal distribution.

        Args:
            loc (_type_): _description_
            chol (_type_): _description_
        """
        super().__init__(loc, chol)

    @property
    def mean(self):
        return self.loc

    @property
    def var(self):
        return self.scale**2

    @property
    def chol_covariance(self):
        return torch.diag_embed(self.scale)

    @property
    def covariance_diagonal(self):
        return self.var

    @property
    def covariance(self):
        return torch.diag_embed(self.var)

    @property
    def precision(self):
        return torch.diag_embed(1.0 / self.var)

    @property
    def logdet_covariance(self):
        return 2 * torch.log(self.scale).sum(-1)

    @property
    def logdet_precision(self):
        return -2 * torch.log(self.scale).sum(-1)

    @property
    def trace_covariance(self):
        return self.var.sum(-1)

    @property
    def trace_precision(self):
        return (1.0 / self.var).sum(-1)

    def covariance_weighted_inner_prod(self, b, reduce_dim=True):
        assert b.shape[-1] == 1
        prod = (self.var.unsqueeze(-1) * (b**2)).sum(-2)
        return prod.squeeze(-1) if reduce_dim else prod

    def precision_weighted_inner_prod(self, b, reduce_dim=True):
        assert b.shape[-1] == 1
        prod = ((b**2) / self.var.unsqueeze(-1)).sum(-2)
        return prod.squeeze(-1) if reduce_dim else prod

    def __add__(self, inp):
        if isinstance(inp, Normal):
            new_cov = self.var + inp.var
            return Normal(
                self.mean + inp.mean, torch.sqrt(torch.clip(new_cov, min=1e-12))
            )
        elif isinstance(inp, torch.Tensor):
            return Normal(self.mean + inp, self.scale)
        else:
            raise NotImplementedError(
                "Distribution addition only implemented for diag covs"
            )

    def __matmul__(self, inp):
        assert inp.shape[-2] == self.loc.shape[-1]
        assert inp.shape[-1] == 1
        new_cov = self.covariance_weighted_inner_prod(
            inp.unsqueeze(-3), reduce_dim=False
        )
        return Normal(self.loc @ inp, torch.sqrt(torch.clip(new_cov, min=1e-12)))

    def squeeze(self, idx):
        return Normal(self.loc.squeeze(idx), self.scale.squeeze(idx))


class DenseNormal(torch.distributions.MultivariateNormal):
    def __init__(self, loc: Tensor, cholesky: Tensor):
        """Dense Normal distribution. Note that this is a multivariate normal used for
        the distribution of the weights in the last layer of a neural network.

        Args:
            loc: Location of the distribution.
            cholesky: Lower triangular Cholesky factor of the covariance matrix.
        """
        super().__init__(loc, scale_tril=cholesky)

    @property
    def mean(self):
        return self.loc

    @property
    def chol_covariance(self):
        return self.scale_tril

    @property
    def covariance(self):
        return self.scale_tril @ tp(self.scale_tril)

    @property
    def inverse_covariance(self):
        logger.warning(
            "Direct matrix inverse for dense covariances is O(N^3),"
            "consider using eg inverse weighted inner product"
        )
        Eye = torch.eye(
            self.scale_tril.shape[-1],
            device=self.scale_tril.device,
            dtype=self.scale_tril.dtype,
        )
        W = torch.linalg.solve_triangular(self.scale_tril, Eye, upper=False)
        return tp(W) @ W

    @property
    def logdet_covariance(self):
        return 2.0 * torch.diagonal(self.scale_tril, dim1=-2, dim2=-1).log().sum(-1)

    @property
    def trace_covariance(self):
        return (self.scale_tril**2).sum(-1).sum(-1)  # compute as frob norm squared

    def covariance_weighted_inner_prod(self, b, reduce_dim=True):
        assert b.shape[-1] == 1
        prod = ((tp(self.scale_tril) @ b) ** 2).sum(-2)
        return prod.squeeze(-1) if reduce_dim else prod

    def precision_weighted_inner_prod(self, b, reduce_dim=True):
        assert b.shape[-1] == 1
        prod = (
            torch.linalg.solve_triangular(self.scale_tril, b, upper=False) ** 2
        ).sum(-2)
        return prod.squeeze(-1) if reduce_dim else prod

    def __matmul__(self, inp):
        assert inp.shape[-2] == self.loc.shape[-1]
        assert inp.shape[-1] == 1
        new_cov = self.covariance_weighted_inner_prod(
            inp.unsqueeze(-3), reduce_dim=False
        )
        return Normal(self.loc @ inp, torch.sqrt(torch.clip(new_cov, min=1e-12)))

    def squeeze(self, idx):
        return DenseNormal(self.loc.squeeze(idx), self.scale_tril.squeeze(idx))


class LowRankNormal(torch.distributions.LowRankMultivariateNormal):
    def __init__(self, loc: Tensor, cov_factor: Tensor, diag: Tensor):
        """Low Rank Normal distribution. Note that this is a multivariate normal used
        for the distribution of the weights in the last layer of a neural network.

        Args:
            loc: Location of the distribution.
            cov_factor: Low rank factor of the covariance matrix.
            diag: Diagonal of the covariance matrix.
        """
        super().__init__(loc, cov_factor=cov_factor, cov_diag=diag)

    @property
    def mean(self):
        return self.loc

    @property
    def chol_covariance(self):
        raise NotImplementedError()

    @property
    def inverse_covariance(self):
        raise NotImplementedError()

    @property
    def logdet_covariance(self):
        # Apply Matrix determinant lemma
        term1 = torch.log(self.cov_diag).sum(-1)
        arg1 = tp(self.cov_factor) @ (self.cov_factor / self.cov_diag.unsqueeze(-1))
        term2 = torch.linalg.det(
            arg1 + torch.eye(arg1.shape[-1], dtype=torch.float64)
        ).log()
        return term1 + term2

    @property
    def trace_covariance(self):
        # trace of sum is sum of traces
        trace_diag = self.cov_diag.sum(-1)
        trace_lowrank = (self.cov_factor**2).sum(-1).sum(-1)
        return trace_diag + trace_lowrank

    def covariance_weighted_inner_prod(self, b, reduce_dim=True):
        assert b.shape[-1] == 1
        diag_term = (self.cov_diag.unsqueeze(-1) * (b**2)).sum(-2)
        factor_term = ((tp(self.cov_factor) @ b) ** 2).sum(-2)
        prod = diag_term + factor_term
        return prod.squeeze(-1) if reduce_dim else prod

    def precision_weighted_inner_prod(self, b, reduce_dim=True):
        raise NotImplementedError()

    def __matmul__(self, inp):
        assert inp.shape[-2] == self.loc.shape[-1]
        assert inp.shape[-1] == 1
        new_cov = self.covariance_weighted_inner_prod(
            inp.unsqueeze(-3), reduce_dim=False
        )
        return Normal(self.loc @ inp, torch.sqrt(torch.clip(new_cov, min=1e-12)))

    def squeeze(self, idx):
        return LowRankNormal(
            self.loc.squeeze(idx),
            self.cov_factor.squeeze(idx),
            self.cov_diag.squeeze(idx),
        )


class DenseNormalPrec(torch.distributions.MultivariateNormal):
    """A DenseNormal parameterized by the mean and the cholesky decomp of the precision
    matrix. Low Rank Normal distribution. Note that this is a multivariate normal used
    for the distribution of the weights in the last layer of a neural network.

    This function also includes a recursive_update function which performs a recursive
    linear regression update with effecient cholesky factor updates.
    """

    def __init__(self, loc: Tensor, cholesky: Tensor, validate_args=False):
        """A DenseNormal parameterized by the mean and the cholesky decomp of the
        precision.

        Args:
            loc: Location of the distribution.
            cholesky: Lower triangular Cholesky factor of the precision matrix.
            validate_args: Whether to validate the input arguments.
        """
        prec = cholesky @ tp(cholesky)
        super().__init__(loc, precision_matrix=prec, validate_args=validate_args)
        self.tril = cholesky

    @property
    def mean(self):
        return self.loc

    @property
    def chol_covariance(self):
        raise NotImplementedError()

    @property
    def covariance(self):
        logger.warning(
            "Direct matrix inverse for dense covariances is O(N^3)"
            "consider using eg inverse weighted inner product"
        )
        return torch.cholesky_inverse(self.tril)

    @property
    def inverse_covariance(self):
        return self.precision_matrix

    @property
    def logdet_covariance(self):
        return -2.0 * torch.diagonal(self.tril, dim1=-2, dim2=-1).log().sum(-1)

    @property
    def trace_covariance(self):
        return (
            (torch.inverse(self.tril) ** 2).sum(-1).sum(-1)
        )  # compute as frob norm squared

    def covariance_weighted_inner_prod(self, b, reduce_dim=True):
        assert b.shape[-1] == 1
        prod = (torch.linalg.solve(self.tril, b) ** 2).sum(-2)
        return prod.squeeze(-1) if reduce_dim else prod

    def precision_weighted_inner_prod(self, b, reduce_dim=True):
        assert b.shape[-1] == 1
        prod = ((tp(self.tril) @ b) ** 2).sum(-2)
        return prod.squeeze(-1) if reduce_dim else prod

    def __matmul__(self, inp):
        assert inp.shape[-2] == self.loc.shape[-1]
        assert inp.shape[-1] == 1
        new_cov = self.covariance_weighted_inner_prod(
            inp.unsqueeze(-3), reduce_dim=False
        )
        return Normal(self.loc @ inp, torch.sqrt(torch.clip(new_cov, min=1e-12)))

    def squeeze(self, idx):
        return DenseNormalPrec(self.loc.squeeze(idx), self.tril.squeeze(idx))


def get_parameterization(p):
    COV_PARAM_DICT = {
        "dense": DenseNormal,
        "dense_precision": DenseNormalPrec,
        "diagonal": Normal,
        "lowrank": LowRankNormal,
    }

    try:
        return COV_PARAM_DICT[p]
    except KeyError:
        raise ValueError(f"Invalid covariance parameterization: {p!r}")


# following functions/classes are from
# https://github.com/VectorInstitute/vbll/blob/main/vbll/layers/regression.py
def gaussian_kl(p, q_scale):
    feat_dim = p.mean.shape[-1]
    mse_term = (p.mean**2).sum(-1).sum(-1) / q_scale
    trace_term = (p.trace_covariance / q_scale).sum(-1)
    logdet_term = (feat_dim * np.log(q_scale) - p.logdet_covariance).sum(-1)
    return 0.5 * (mse_term + trace_term + logdet_term)  # currently exclude constant


@dataclass
class VBLLReturn:
    predictive: Normal | DenseNormal
    train_loss_fn: Callable[[torch.Tensor], torch.Tensor]
    val_loss_fn: Callable[[torch.Tensor], torch.Tensor]


class Regression(nn.Module):
    def __init__(
        self,
        in_features,
        out_features,
        regularization_weight,
        parameterization="dense",
        prior_scale=1.0,
        wishart_scale=1e-2,
        cov_rank=None,
        clamp_noise_init=True,
        dof=1.0,
    ):
        """
        Variational Bayesian Linear Regression

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        regularization_weight : float
            Weight on regularization term in ELBO
        parameterization : str
            Parameterization of covariance matrix.
            Currently supports {'dense', 'diagonal', 'lowrank', 'dense_precision'}
        prior_scale : float
            Scale of prior covariance matrix
        wishart_scale : float
            Scale of Wishart prior on noise covariance
        dof : float
            Degrees of freedom of Wishart prior on noise covariance
        """
        super().__init__()

        self.wishart_scale = wishart_scale
        self.dof = (dof + out_features + 1.0) / 2.0
        self.regularization_weight = regularization_weight
        self.dtype = torch.float64  # NOTE: not in the original source code

        # define prior, currently fixing zero mean and arbitrarily scaled cov
        self.prior_scale = prior_scale * (1.0 / in_features)

        # noise distribution
        self.noise_mean = nn.Parameter(
            torch.zeros(out_features, dtype=self.dtype), requires_grad=False
        )
        self.noise_logdiag = nn.Parameter(
            torch.randn(out_features, dtype=self.dtype) * (np.log(wishart_scale))
        )

        # ensure that log noise is positive
        if clamp_noise_init:
            self.noise_logdiag.data = torch.clamp(self.noise_logdiag.data, min=0)

        # last layer distribution
        self.W_dist = get_parameterization(parameterization)
        self.W_mean = nn.Parameter(
            torch.randn(out_features, in_features, dtype=self.dtype)
        )

        if parameterization == "diagonal":
            self.W_logdiag = nn.Parameter(
                torch.randn(out_features, in_features, dtype=self.dtype)
                - 0.5 * np.log(in_features)
            )
        elif parameterization == "dense":
            self.W_logdiag = nn.Parameter(
                torch.randn(out_features, in_features, dtype=self.dtype)
                - 0.5 * np.log(in_features)
            )
            self.W_offdiag = nn.Parameter(
                torch.randn(out_features, in_features, in_features, dtype=self.dtype)
                / in_features
            )
        elif parameterization == "dense_precision":
            self.W_logdiag = nn.Parameter(
                torch.randn(out_features, in_features, dtype=self.dtype)
                + 0.5 * np.log(in_features)
            )
            self.W_offdiag = nn.Parameter(
                torch.randn(out_features, in_features, in_features, dtype=self.dtype)
                * 0.0
            )
        elif parameterization == "lowrank":
            if cov_rank is None:
                raise ValueError("Must specify cov_rank for lowrank parameterization")

            self.W_logdiag = nn.Parameter(
                torch.randn(out_features, in_features, dtype=self.dtype)
                - 0.5 * np.log(in_features)
            )
            self.W_offdiag = nn.Parameter(
                torch.randn(out_features, in_features, cov_rank, dtype=self.dtype)
                / in_features
            )

    def W(self):
        cov_diag = torch.exp(self.W_logdiag)
        if self.W_dist == Normal:
            cov = self.W_dist(self.W_mean, cov_diag)
        elif (self.W_dist == DenseNormal) or (self.W_dist == DenseNormalPrec):
            tril = torch.tril(self.W_offdiag, diagonal=-1) + torch.diag_embed(cov_diag)
            cov = self.W_dist(self.W_mean, tril)
        elif self.W_dist == LowRankNormal:
            cov = self.W_dist(self.W_mean, self.W_offdiag, cov_diag)

        return cov

    def noise(self):
        return Normal(self.noise_mean, torch.exp(self.noise_logdiag))

    def forward(self, x):
        out = VBLLReturn(
            self.predictive(x), self._get_train_loss_fn(x), self._get_val_loss_fn(x)
        )
        return out

    def predictive(self, x):
        return (self.W() @ x[..., None]).squeeze(-1) + self.noise()

    def _get_train_loss_fn(self, x):
        def loss_fn(y):
            # construct predictive density N(W @ phi, Sigma)
            W = self.W()
            noise = self.noise()
            pred_density = Normal((W.mean @ x[..., None]).squeeze(-1), noise.scale)
            pred_likelihood = pred_density.log_prob(y)

            trace_term = 0.5 * (
                (W.covariance_weighted_inner_prod(x.unsqueeze(-2)[..., None]))
                * noise.trace_precision
            )

            kl_term = gaussian_kl(W, self.prior_scale)
            wishart_term = (
                self.dof * noise.logdet_precision
                - 0.5 * self.wishart_scale * noise.trace_precision
            )
            total_elbo = torch.mean(pred_likelihood - trace_term)
            regularization_term = self.regularization_weight * (wishart_term - kl_term)
            total_elbo = total_elbo + regularization_term
            return -total_elbo

        return loss_fn

    def _get_val_loss_fn(self, x):
        def loss_fn(y):
            # compute log likelihood under variational posterior via marginalization
            logprob = self.predictive(x).log_prob(y).sum(-1)  # sum over output dims
            return -logprob.mean(0)  # mean over batch dim

        return loss_fn



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
A `SamplerList` for sampling from a `PosteriorList`.
"""

from __future__ import annotations

import torch
from botorch.exceptions.errors import UnsupportedError
from botorch.posteriors.posterior_list import PosteriorList
from botorch.sampling.base import MCSampler
from torch import Tensor
from torch.nn import ModuleList


class ListSampler(MCSampler):
    def __init__(self, *samplers: MCSampler) -> None:
        r"""A list of samplers for sampling from a `PosteriorList`.

        Args:
            samplers: A variable number of samplers. This should include
                a sampler for each posterior.
        """
        super(MCSampler, self).__init__()
        self.samplers = ModuleList(samplers)
        self._validate_samplers()

    def _validate_samplers(self) -> None:
        r"""Checks that the samplers share the same sample shape."""
        sample_shapes = [s.sample_shape for s in self.samplers]
        if not all(sample_shapes[0] == ss for ss in sample_shapes):
            raise UnsupportedError(
                "ListSampler requires all samplers to have the same sample shape."
            )

    @property
    def sample_shape(self) -> torch.Size:
        r"""The sample shape of the underlying samplers."""
        self._validate_samplers()
        return self.samplers[0].sample_shape

    def forward(self, posterior: PosteriorList) -> Tensor:
        r"""Samples from the posteriors and concatenates the samples.

        Args:
            posterior: A `PosteriorList` to sample from.

        Returns:
            The samples drawn from the posterior.
        """
        samples_list = [
            s(posterior=p) for s, p in zip(self.samplers, posterior.posteriors)
        ]
        return posterior._reshape_and_cat(tensors=samples_list)

    def _update_base_samples(
        self, posterior: PosteriorList, base_sampler: ListSampler
    ) -> None:
        r"""Update the sampler to use the original base samples for X_baseline.

        This is used in CachedCholeskyAcquisitionFunctions to ensure consistency.

        Args:
            posterior: The posterior for which the base samples are constructed.
            base_sampler: The base sampler to retrieve the base samples from.
        """
        self._instance_check(base_sampler=base_sampler)
        for s, p, bs in zip(self.samplers, posterior.posteriors, base_sampler.samplers):
            s._update_base_samples(posterior=p, base_sampler=bs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import math
from collections.abc import Callable
from dataclasses import dataclass

import torch
from torch import Tensor
from torch.nn import Parameter


class FixedHomotopySchedule:
    """Homotopy schedule with a fixed list of values."""

    def __init__(self, values: list[float]) -> None:
        r"""Initialize FixedHomotopySchedule.

        Args:
            values: A list of values used in homotopy
        """
        self._values = values
        self.idx = 0

    @property
    def num_steps(self) -> int:
        return len(self._values)

    @property
    def value(self) -> float:
        return self._values[self.idx]

    @property
    def should_stop(self) -> bool:
        return self.idx == len(self._values)

    def restart(self) -> None:
        self.idx = 0

    def step(self) -> None:
        self.idx += 1


class LinearHomotopySchedule(FixedHomotopySchedule):
    """Linear homotopy schedule."""

    def __init__(self, start: float, end: float, num_steps: int) -> None:
        r"""Initialize LinearHomotopySchedule.

        Args:
            start: start value of homotopy
            end: end value of homotopy
            num_steps: number of steps in the homotopy schedule.
        """
        super().__init__(
            values=torch.linspace(start, end, num_steps, dtype=torch.double).tolist()
        )


class LogLinearHomotopySchedule(FixedHomotopySchedule):
    """Log-linear homotopy schedule."""

    def __init__(self, start: float, end: float, num_steps: int):
        r"""Initialize LogLinearHomotopySchedule.

        Args:
            start: start value of homotopy
            end: end value of homotopy
            num_steps: number of steps in the homotopy schedule.
        """
        super().__init__(
            values=torch.logspace(
                math.log10(start), math.log10(end), num_steps, dtype=torch.double
            ).tolist()
        )


@dataclass
class HomotopyParameter:
    r"""Homotopy parameter.

    The parameter is expected to either be a torch parameter or a torch tensor which may
    correspond to a buffer of a module. The parameter has a corresponding schedule.
    """

    parameter: Parameter | Tensor
    schedule: FixedHomotopySchedule


class Homotopy:
    """Generic homotopy class.

    This class is designed to be used in `optimize_acqf_homotopy`. Given a set of
    homotopy parameters and corresponding schedules we step through the homotopies
    until we have solved the final problem. We additionally support passing in a list
    of callbacks that will be executed each time `step`, `reset`, and `restart` are
    called.
    """

    def __init__(
        self,
        homotopy_parameters: list[HomotopyParameter],
        callbacks: list[Callable] | None = None,
    ) -> None:
        r"""Initialize the homotopy.

        Args:
            homotopy_parameters: List of homotopy parameters
            callbacks: Optional list of callbacks that are executed each time
                `restart`, `reset`, or `step` are called. These may be used to, e.g.,
                reinitialize the acquisition function which is needed when using qNEHVI.
        """
        self._homotopy_parameters = homotopy_parameters
        self._callbacks = callbacks or []
        self._original_values = [
            hp.parameter.item() for hp in self._homotopy_parameters
        ]
        assert all(
            isinstance(hp.parameter, Parameter) or isinstance(hp.parameter, Tensor)
            for hp in self._homotopy_parameters
        )
        # Assume the same number of steps for now
        assert len({h.schedule.num_steps for h in self._homotopy_parameters}) == 1
        # Initialize the homotopy parameters
        self.restart()

    def _execute_callbacks(self) -> None:
        """Execute the callbacks."""
        for callback in self._callbacks:
            callback()

    @property
    def should_stop(self) -> bool:
        """Returns true if all schedules have reached the end."""
        return all(h.schedule.should_stop for h in self._homotopy_parameters)

    def restart(self) -> None:
        """Restart the homotopy to use the initial value in the schedule."""
        for hp in self._homotopy_parameters:
            hp.schedule.restart()
            hp.parameter.data.fill_(hp.schedule.value)
        self._execute_callbacks()

    def reset(self) -> None:
        """Reset the homotopy parameter to their original values."""
        for hp, val in zip(self._homotopy_parameters, self._original_values):
            hp.parameter.data.fill_(val)
        self._execute_callbacks()

    def step(self) -> None:
        """Take a step according to the schedules."""
        for hp in self._homotopy_parameters:
            hp.schedule.step()
            if not hp.schedule.should_stop:
                hp.parameter.data.fill_(hp.schedule.value)
        self._execute_callbacks()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import math
from collections.abc import Callable, Iterable, Iterator

from functools import lru_cache
from math import pi
from numbers import Number
from typing import Any

import torch
from botorch.utils.safe_math import logdiffexp
from numpy.polynomial.legendre import leggauss as numpy_leggauss
from torch import BoolTensor, LongTensor, Tensor

CaseNd = tuple[Callable[[], BoolTensor], Callable[[BoolTensor], Tensor]]

_log_2 = math.log(2)
_sqrt_pi = math.sqrt(pi)
_inv_sqrt_pi = 1 / _sqrt_pi
_inv_sqrt_2pi = 1 / math.sqrt(2 * pi)
_inv_sqrt_2 = 1 / math.sqrt(2)
_neg_inv_sqrt_2 = -_inv_sqrt_2
_log_sqrt_2pi = math.log(2 * pi) / 2
STANDARDIZED_RANGE: tuple[float, float] = (-1e6, 1e6)
_log_two_inv_sqrt_2pi = _log_2 - _log_sqrt_2pi  # = log(2 / sqrt(2 * pi))


def case_dispatcher(
    out: Tensor,
    cases: Iterable[CaseNd] = (),
    default: Callable[[BoolTensor], Tensor] = None,
) -> Tensor:
    r"""Basic implementation of a tensorized switching case statement.

    Args:
        out: Tensor to which case outcomes are written.
        cases: Iterable of function pairs (pred, func), where `mask=pred()` specifies
            whether `func` is applicable for each entry in `out`. Note that cases are
            resolved first-come, first-serve.
        default: Optional `func` to which all unclaimed entries of `out` are dispatched.
    """
    active = None
    for closure, func in cases:
        pred = closure()
        if not pred.any():
            continue

        mask = pred if (active is None) else pred & active
        if not mask.any():
            continue

        if mask.all():  # where possible, use Ellipsis to avoid indexing
            out[...] = func(...)
            return out

        out[mask] = func(mask)
        if active is None:
            active = ~mask
        else:
            active[mask] = False

        if not active.any():
            break

    if default is not None:
        if active is None:
            out[...] = default(...)
        elif active.any():
            out[active] = default(active)

    return out


@lru_cache(maxsize=None)
def get_constants(
    values: Number | Iterator[Number],
    device: torch.device | None = None,
    dtype: torch.dtype | None = None,
) -> Tensor | tuple[Tensor, ...]:
    r"""Returns scalar-valued Tensors containing each of the given constants.
    Used to expedite tensor operations involving scalar arithmetic. Note that
    the returned Tensors should not be modified in-place."""
    if isinstance(values, Number):
        return torch.full((), values, dtype=dtype, device=device)

    return tuple(torch.full((), val, dtype=dtype, device=device) for val in values)


def get_constants_like(
    values: Number | Iterator[Number],
    ref: Tensor,
) -> Tensor | Iterator[Tensor]:
    return get_constants(values, device=ref.device, dtype=ref.dtype)


def gen_positional_indices(
    shape: torch.Size,
    dim: int,
    device: torch.device | None = None,
) -> Iterator[torch.LongTensor]:
    ndim = len(shape)
    _dim = ndim + dim if dim < 0 else dim
    if _dim >= ndim or _dim < 0:
        raise ValueError(f"dim={dim} invalid for shape {shape}.")

    cumsize = shape[_dim + 1 :].numel()
    for i, s in enumerate(reversed(shape[: _dim + 1])):
        yield torch.arange(0, s * cumsize, cumsize, device=device)[(...,) + i * (None,)]
        cumsize *= s


def build_positional_indices(
    shape: torch.Size,
    dim: int,
    device: torch.device | None = None,
) -> LongTensor:
    return sum(gen_positional_indices(shape=shape, dim=dim, device=device))


@lru_cache(maxsize=None)
def leggauss(deg: int, **tkwargs: Any) -> tuple[Tensor, Tensor]:
    x, w = numpy_leggauss(deg)
    return torch.as_tensor(x, **tkwargs), torch.as_tensor(w, **tkwargs)


def ndtr(x: Tensor) -> Tensor:
    r"""Standard normal CDF."""
    half, neg_inv_sqrt_2 = get_constants_like((0.5, _neg_inv_sqrt_2), x)
    return half * torch.erfc(neg_inv_sqrt_2 * x)


def phi(x: Tensor) -> Tensor:
    r"""Standard normal PDF."""
    inv_sqrt_2pi, neg_half = get_constants_like((_inv_sqrt_2pi, -0.5), x)
    return inv_sqrt_2pi * (neg_half * x.square()).exp()


def log_phi(x: Tensor) -> Tensor:
    r"""Logarithm of standard normal pdf"""
    log_sqrt_2pi, neg_half = get_constants_like((_log_sqrt_2pi, -0.5), x)
    return neg_half * x.square() - log_sqrt_2pi


def log_ndtr(x: Tensor) -> Tensor:
    """Implementation of log_ndtr that remedies problems of torch.special's version
    for large negative x, where the torch implementation yields Inf or NaN gradients.

    Args:
        x: An input tensor with dtype torch.float32 or torch.float64.

    Returns:
        A tensor of values of the same type and shape as x containing log(ndtr(x)).
    """
    if not (x.dtype == torch.float32 or x.dtype == torch.float64):
        raise TypeError(
            f"log_Phi only supports torch.float32 and torch.float64 "
            f"dtypes, but received {x.dtype=}."
        )
    neg_inv_sqrt_2, log_2 = get_constants_like((_neg_inv_sqrt_2, _log_2), x)
    return log_erfc(neg_inv_sqrt_2 * x) - log_2


def log_erfc(x: Tensor) -> Tensor:
    """Computes the logarithm of the complementary error function in a numerically
    stable manner. The GitHub issue https://github.com/pytorch/pytorch/issues/31945
    tracks progress toward moving this feature into PyTorch in C++.

    Args:
        x: An input tensor with dtype torch.float32 or torch.float64.

    Returns:
        A tensor of values of the same type and shape as x containing log(erfc(x)).
    """
    if not (x.dtype == torch.float32 or x.dtype == torch.float64):
        raise TypeError(
            f"log_erfc only supports torch.float32 and torch.float64 "
            f"dtypes, but received {x.dtype=}."
        )
    is_pos = x > 0
    x_pos = x.masked_fill(~is_pos, 0)
    x_neg = x.masked_fill(is_pos, 0)
    return torch.where(
        is_pos,
        torch.log(torch.special.erfcx(x_pos)) - x_pos.square(),
        torch.log(torch.special.erfc(x_neg)),
    )


def log_erfcx(x: Tensor) -> Tensor:
    """Computes the logarithm of the complementary scaled error function in a
    numerically stable manner. The GitHub issue tracks progress toward moving this
    feature into PyTorch in C++: https://github.com/pytorch/pytorch/issues/31945.

    Args:
        x: An input tensor with dtype torch.float32 or torch.float64.

    Returns:
        A tensor of values of the same type and shape as x containing log(erfcx(x)).
    """
    is_pos = x > 0
    x_pos = x.masked_fill(~is_pos, 0)
    x_neg = x.masked_fill(is_pos, 0)
    return torch.where(
        is_pos,
        torch.special.erfcx(x_pos).log(),
        torch.special.erfc(x_neg).log() + x.square(),
    )


def standard_normal_log_hazard(x: Tensor) -> Tensor:
    """Computes the logarithm of the hazard function of the standard normal
    distribution, i.e. `log(phi(x) / Phi(-x))`.

    Args:
        x: A tensor of any shape, with either float32 or float64 dtypes.

    Returns:
        A Tensor of the same shape `x`, containing the values of the logarithm of the
        hazard function evaluated at `x`.
    """
    # NOTE: using _inv_sqrt_2 instead of _neg_inv_sqrt_2 means we are computing Phi(-x).
    a, b = get_constants_like((_log_two_inv_sqrt_2pi, _inv_sqrt_2), x)
    return a - log_erfcx(b * x)


def log_prob_normal_in(a: Tensor, b: Tensor) -> Tensor:
    r"""Computes the probability that a standard normal random variable takes a value
    in \[a, b\], i.e. log(Phi(b) - Phi(a)), where Phi is the standard normal CDF.
    Returns accurate values and permits numerically stable backward passes for inputs
    in [-1e100, 1e100] for double precision and [-1e20, 1e20] for single precision.
    In contrast, a naive approach is not numerically accurate beyond [-10, 10].

    Args:
        a: Tensor of lower integration bounds of the Gaussian probability measure.
        b: Tensor of upper integration bounds of the Gaussian probability measure.

    Returns:
        Tensor of the log probabilities.
    """
    if not (a < b).all():
        raise ValueError("Received input tensors a, b for which not all a < b.")
    # if abs(b) > abs(a), we use Phi(b) - Phi(a) = Phi(-a) - Phi(-b), since the
    # right tail converges to 0 from below, leading to digit cancellation issues, while
    # the left tail of log_ndtr is well behaved and results in large negative numbers
    rev_cond = b.abs() > a.abs()  # condition for reversal of inputs
    if rev_cond.any():
        c = torch.where(rev_cond, -b, a)
        b = torch.where(rev_cond, -a, b)
        a = c  # after we updated b, can assign c to a
    return logdiffexp(log_a=log_ndtr(a), log_b=log_ndtr(b))


def swap_along_dim_(
    values: Tensor,
    i: int | LongTensor,
    j: int | LongTensor,
    dim: int,
    buffer: Tensor | None = None,
) -> Tensor:
    r"""Swaps Tensor slices in-place along dimension `dim`.

    When passed as Tensors, `i` (and `j`) should be `dim`-dimensional tensors
    with the same shape as `values.shape[:dim]`. The xception to this rule occurs
    when `dim=0`, in which case `i` (and `j`) should be (at most) one-dimensional
    when passed as a Tensor.

    Args:
        values: Tensor whose values are to be swapped.
        i: Indices for slices along dimension `dim`.
        j: Indices for slices along dimension `dim`.
        dim: The dimension of `values` along which to swap slices.
        buffer: Optional buffer used internally to store copied values.

    Returns:
        The original `values` tensor.
    """
    dim = values.ndim + dim if dim < 0 else dim
    if dim and (
        (isinstance(i, Tensor) and i.ndim) or (isinstance(j, Tensor) and j.ndim)
    ):
        # Handle n-dimensional batches of heterogeneous swaps via linear indexing
        if isinstance(i, Tensor) and i.shape != values.shape[:dim]:
            raise ValueError("Batch shapes of `i` and `values` do not match.")

        if isinstance(j, Tensor) and j.shape != values.shape[:dim]:
            raise ValueError("Batch shapes of `j` and `values` do not match.")

        pidx = build_positional_indices(
            shape=values.shape[: dim + 1], dim=-2, device=values.device
        )

        swap_along_dim_(
            values.view(-1, *values.shape[dim + 1 :]),
            i=(pidx + i).view(-1),
            j=(pidx + j).view(-1),
            dim=0,
            buffer=buffer,
        )

    else:
        # Base cases: homogeneous swaps and 1-dimenensional heterogeneous swaps
        if isinstance(i, Tensor) and i.ndim > 1:
            raise ValueError("Tensor `i` must be at most 1-dimensional when `dim=0`.")

        if isinstance(j, Tensor) and j.ndim > 1:
            raise ValueError("Tensor `j` must be at most 1-dimensional when `dim=0`.")

        if dim:
            ctx = tuple(slice(None) for _ in range(dim))
            i = ctx + (i,)
            j = ctx + (j,)

        if buffer is None:
            buffer = values[i].clone()
        else:
            buffer.copy_(values[i])

        values[i] = values[j]
        values[j] = buffer

    return values


def compute_log_prob_feas_from_bounds(
    con_lower_inds: Tensor,
    con_upper_inds: Tensor,
    con_both_inds: Tensor,
    con_lower: Tensor,
    con_upper: Tensor,
    con_both: Tensor,
    means: Tensor,
    sigmas: Tensor,
) -> Tensor:
    r"""Compute logarithm of the feasibility probability for each batch of mean/sigma.

    Args:
        means: A `(b) x m`-dim Tensor of means.
        sigmas: A `(b) x m`-dim Tensor of standard deviations.
        con_lower_inds: 1d Tensor of indices con_lower applies to
            in the second dimension of means and sigmas.
        con_upper_inds: 1d Tensor of indices con_upper applies to
            in the second dimension of means and sigmas.
        con_both_inds: 1d Tensor of indices con_both applies to
            in the second dimension of means and sigmas.
        con_lower: 1d Tensor of lower bounds on the constraints
            equal in dimension to con_lower_inds.
        con_upper: 1d Tensor of upper bounds on the constraints
            equal in dimension to con_upper_inds.
        con_both: 2d Tensor of "both" bounds on the constraints
            equal in length to con_both_inds.

    Returns:
        A `b`-dim tensor of log feasibility probabilities
    """
    # indices are integers, so we don't cast the type
    con_upper_inds = con_upper_inds.to(device=means.device)
    con_lower_inds = con_lower_inds.to(device=means.device)
    con_both_inds = con_both_inds.to(device=means.device)

    con_lower = con_lower.to(device=means.device, dtype=means.dtype)
    con_upper = con_upper.to(device=means.device, dtype=means.dtype)
    con_both = con_both.to(device=means.device, dtype=means.dtype)

    log_prob = torch.zeros_like(means[..., 0])
    if len(con_lower_inds) > 0:
        i = con_lower_inds
        dist_l = (con_lower - means[..., i]) / sigmas[..., i]
        log_prob = log_prob + log_ndtr(-dist_l).sum(dim=-1)  # 1 - Phi(x) = Phi(-x)
    if len(con_upper_inds) > 0:
        i = con_upper_inds
        dist_u = (con_upper - means[..., i]) / sigmas[..., i]
        log_prob = log_prob + log_ndtr(dist_u).sum(dim=-1)
    if len(con_both_inds) > 0:
        i = con_both_inds
        con_lower, con_upper = con_both[:, 0], con_both[:, 1]
        # scaled distance to lower and upper constraint boundary:
        dist_l = (con_lower - means[..., i]) / sigmas[..., i]
        dist_u = (con_upper - means[..., i]) / sigmas[..., i]
        log_prob = log_prob + log_prob_normal_in(a=dist_l, b=dist_u).sum(dim=-1)
    return log_prob


def percentile_of_score(data: Tensor, score: Tensor, dim: int = -1) -> Tensor:
    """Compute the percentile rank of `score` relative to `data`.
    For example, if this function returns 70 then 70% of the
    values in `data` are below `score`.

    This implementation is based on `scipy.stats.percentileofscore`,
    with `kind='rank'` and `nan_policy='propagate'`, which is the default.

    Args:
        data: A `... x n x output_shape`-dim Tensor of data.
        score: A `... x 1 x output_shape`-dim Tensor of scores.

    Returns:
        A `... x output_shape`-dim Tensor of percentile ranks.
    """
    # based on scipy.stats.percentileofscore
    left = torch.count_nonzero(data < score, dim=dim)
    right = torch.count_nonzero(data <= score, dim=dim)
    plus1 = left < right
    perct = (left + right + plus1) * (50.0 / data.shape[dim])
    # perct shape: `... x output_shape`
    # fill in nans due to current trial progression being nan
    nan_mask = torch.broadcast_to(torch.isnan(score.squeeze(dim)), perct.shape)
    perct[nan_mask] = torch.nan
    # fill in nans due to previous trial progressions being nan
    nan_mask = torch.broadcast_to(torch.any(torch.isnan(data), dim=dim), perct.shape)
    perct[nan_mask] = torch.nan
    return perct



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Preference acquisition functions. This includes:
Analytical EUBO acquisition function as introduced in [Lin2022preference]_
and its MC-based generalization qEUBO as proposed in [Astudillo2023qeubo]_.

.. [Astudillo2023qeubo]
    Astudillo, R., Lin, Z.J., Bakshy, E. and Frazier, P.I. qEUBO: A Decision-Theoretic
    Acquisition Function for Preferential Bayesian Optimization. International
    Conference on Artificial Intelligence and Statistics (AISTATS), 2023.

.. [Lin2022preference]
    Lin, Z.J., Astudillo, R., Frazier, P.I. and Bakshy, E. Preference Exploration
    for Efficient Bayesian Optimization with Multiple Outcomes. International
    Conference on Artificial Intelligence and Statistics (AISTATS), 2022.

.. [Houlsby2011bald]
    Houlsby, N., Huszár, F., Ghahramani, Z. and Lengyel, M.
    Bayesian Active Learning for Gaussian Process Classification.
    NIPS Workshop on Bayesian optimization, experimental design and bandits:
    Theory and applications, 2011.
"""

from __future__ import annotations

import torch
from botorch.acquisition import AnalyticAcquisitionFunction
from botorch.acquisition.monte_carlo import MCAcquisitionFunction
from botorch.acquisition.objective import MCAcquisitionObjective, PosteriorTransform
from botorch.exceptions.errors import UnsupportedError
from botorch.models.deterministic import DeterministicModel
from botorch.models.model import Model
from botorch.sampling.base import MCSampler
from botorch.utils.transforms import (
    average_over_ensemble_models,
    concatenate_pending_points,
    match_batch_shape,
    t_batch_mode_transform,
)
from torch import Tensor
from torch.distributions import Bernoulli, Normal

SIGMA_JITTER = 1e-8


class AnalyticExpectedUtilityOfBestOption(AnalyticAcquisitionFunction):
    r"""Analytic Preferential Expected Utility of Best Options, i.e., Analytical EUBO"""

    def __init__(
        self,
        pref_model: Model,
        outcome_model: DeterministicModel | None = None,
        previous_winner: Tensor | None = None,
    ) -> None:
        r"""Analytic implementation of Expected Utility of the Best Option under the
        Laplace model (assumes a PairwiseGP is used as the preference model) as
        proposed in [Lin2022preference]_.

        Args:
            pref_model: The preference model that maps the outcomes (i.e., Y) to
                scalar-valued utility.
            outcome_model: A deterministic model that maps parameters (i.e., X) to
                outcomes (i.e., Y). The outcome model f defines the search space of
                Y = f(X). If model is None, we are directly calculating EUBO on
                the parameter space. When used with `OneSamplePosteriorDrawModel`,
                we are obtaining EUBO-zeta as described in [Lin2022preference]_.
            previous_winner: Tensor representing the previous winner in the Y space.
        """
        super().__init__(model=pref_model)
        # ensure the model is in eval mode
        self.add_module("outcome_model", outcome_model)
        self.register_buffer("previous_winner", previous_winner)

        tkwargs = {
            "dtype": pref_model.datapoints.dtype,
            "device": pref_model.datapoints.device,
        }
        std_norm = torch.distributions.normal.Normal(
            torch.zeros(1, **tkwargs),
            torch.ones(1, **tkwargs),
        )
        self.std_norm = std_norm

    @t_batch_mode_transform()
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate analytical EUBO on the candidate set X.

        Args:
            X: A `batch_shape x q x d`-dim Tensor, where `q = 2` if `previous_winner`
                is not `None`, and `q = 1` otherwise.

        Returns:
            The acquisition value for each batch as a tensor of shape `batch_shape`.
        """
        if not (
            ((X.shape[-2] == 2) and (self.previous_winner is None))
            or ((X.shape[-2] == 1) and (self.previous_winner is not None))
        ):
            raise UnsupportedError(
                f"{self.__class__.__name__} only support q=2 or q=1"
                "with a previous winner specified"
            )

        Y = X if self.outcome_model is None else self.outcome_model(X)

        if self.previous_winner is not None:
            Y = torch.cat([Y, match_batch_shape(self.previous_winner, Y)], dim=-2)

        pref_posterior = self.model.posterior(Y)
        pref_mean = pref_posterior.mean.squeeze(-1)
        pref_cov = pref_posterior.covariance_matrix
        delta = pref_mean[..., 0] - pref_mean[..., 1]

        w = torch.tensor([1.0, -1.0], dtype=pref_cov.dtype, device=pref_cov.device)
        var = w @ pref_cov @ w
        sigma = torch.sqrt(var.clamp(min=SIGMA_JITTER))

        u = delta / sigma

        ucdf = self.std_norm.cdf(u)
        updf = torch.exp(self.std_norm.log_prob(u))
        acqf_val = sigma * (updf + u * ucdf)
        if self.previous_winner is None:
            acqf_val = acqf_val + pref_mean[..., 1]
        return acqf_val


class qExpectedUtilityOfBestOption(MCAcquisitionFunction):
    r"""MC-based Expected Utility of Best Option (qEUBO)

    This computes qEUBO by
    (1) sampling the joint posterior over q points
    (2) evaluating the maximum objective value accross the q points
    (3) averaging over the samples

    `qEUBO(X) = E[max Y], Y ~ f(X), where X = (x_1,...,x_q)`
    """

    def __init__(
        self,
        pref_model: Model,
        outcome_model: DeterministicModel | None = None,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
    ) -> None:
        r"""MC-based Expected Utility of Best Option (qEUBO) as proposed
        in [Astudillo2023qeubo]_.

        Args:
            pref_model: The preference model that maps the outcomes (i.e., Y) to
                scalar-valued utility.
            outcome_model: A deterministic model that maps parameters (i.e., X) to
                outcomes (i.e., Y). The outcome model f defines the search space of
                Y = f(X). If model is None, we are directly calculating qEUBO on
                the parameter space.
             sampler: The sampler used to draw base samples. See `MCAcquisitionFunction`
                more details.
            objective: The MCAcquisitionObjective under which the samples are evaluated.
                Defaults to `IdentityMCObjective()`.
            posterior_transform: A PosteriorTransform (optional).
            X_pending:  A `m x d`-dim Tensor of `m` design points that have been
                submitted for function evaluation but have not yet been evaluated.
                Concatenated into X upon forward call. Copied and set
                to have no gradient.
        """
        super().__init__(
            model=pref_model,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            X_pending=X_pending,
        )
        # ensure the model is in eval mode
        self.add_module("outcome_model", outcome_model)

    @concatenate_pending_points
    @t_batch_mode_transform()
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate qEUBO on the candidate set `X`.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of t-batches with `q`
                `d`-dim design points each.

        Returns:
            A `batch_shape'`-dim Tensor of qEUBO values at the given design
            points `X`, where `batch_shape'` is the broadcasted batch shape
            of model and input `X`.
        """
        Y = X if self.outcome_model is None else self.outcome_model(X)

        _, obj = self._get_samples_and_objectives(Y)
        obj_best = obj.max(dim=-1).values
        return obj_best.mean(dim=0)


class PairwiseBayesianActiveLearningByDisagreement(MCAcquisitionFunction):
    r"""MC Bayesian Active Learning by Disagreement"""

    def __init__(
        self,
        pref_model: Model,
        outcome_model: DeterministicModel | None = None,
        num_samples: int | None = 1024,
        std_noise: float | None = 0.0,
    ) -> None:
        """
        Monte Carlo implementation of Bayesian Active Learning by Disagreement (BALD)
        proposed in [Houlsby2011bald]_.

        Args:
            pref_model: The preference model that maps the outcomes (i.e., Y) to
                scalar-valued utility.
            outcome_model: A deterministic model that maps parameters (i.e., X) to
                outcomes (i.e., Y). The outcome model f defines the search space of
                Y = f(X). If model is None, we are directly calculating BALD on
                the parameter space.
            num_samples: number of samples to approximate the conditional_entropy.
            std_noise: Additional observational noise to include. Defaults to 0.
        """
        super().__init__(model=pref_model)
        # ensure the model is in eval mode
        self.add_module("outcome_model", outcome_model)

        self.num_samples = num_samples
        # assuming the relative observation noise is fixed at 1.0 (e.g., in PairwiseGP)
        self.std_noise = std_noise
        self.std_normal = Normal(0, 1)

    @t_batch_mode_transform(expected_q=2)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate MC BALD on the candidate set `X`.

        Args:
            X: A `batch_shape x 2 x d`-dim Tensor of t-batches with `q=2`
                `d`-dim design points each.

        Returns:
            A `batch_shape'`-dim Tensor of MC BALD values at the given
            design points pair `X`, where `batch_shape'` is the broadcasted
            batch shape of model and input `X`.
        """
        Y = X if self.outcome_model is None else self.outcome_model(X)

        pref_posterior = self.model.posterior(Y)
        pref_mean = pref_posterior.mean.squeeze(-1)
        pref_cov = pref_posterior.covariance_matrix

        mu = pref_mean[..., 0] - pref_mean[..., 1]
        w = torch.tensor([1.0, -1.0], dtype=pref_cov.dtype, device=pref_cov.device)
        var = 2 * self.std_noise + w @ pref_cov @ w
        sigma = torch.sqrt(var.clamp(min=SIGMA_JITTER))

        # eq (3) in Houlsby, et al. (2011)
        posterior_entropies = Bernoulli(
            self.std_normal.cdf(mu / torch.sqrt(var + 1))
        ).entropy()

        # Sample-based approx to eq (4) in Houlsby, et al. (2011)
        obj_samples = self.std_normal.cdf(
            Normal(loc=mu, scale=sigma).rsample(torch.Size([self.num_samples]))
        )
        sample_entropies = Bernoulli(obj_samples).entropy()
        conditional_entropies = sample_entropies.mean(dim=0)

        return posterior_entropies - conditional_entropies



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from collections.abc import Iterator

from functools import lru_cache
from numbers import Number

import torch
from torch import Tensor


@lru_cache(maxsize=None)
def get_constants(
    values: Number | Iterator[Number],
    device: torch.device | None = None,
    dtype: torch.dtype | None = None,
) -> Tensor | tuple[Tensor, ...]:
    r"""Returns scalar-valued Tensors containing each of the given constants.
    Used to expedite tensor operations involving scalar arithmetic. Note that
    the returned Tensors should not be modified in-place."""
    if isinstance(values, Number):
        return torch.full((), values, dtype=dtype, device=device)

    return tuple(torch.full((), val, dtype=dtype, device=device) for val in values)


def get_constants_like(
    values: Number | Iterator[Number],
    ref: Tensor,
) -> Tensor | Iterator[Tensor]:
    return get_constants(values, device=ref.device, dtype=ref.dtype)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from botorch.utils.probability.bvn import bvn, bvnmom
from botorch.utils.probability.lin_ess import LinearEllipticalSliceSampler
from botorch.utils.probability.mvnxpb import MVNXPB
from botorch.utils.probability.truncated_multivariate_normal import (
    TruncatedMultivariateNormal,
)
from botorch.utils.probability.unified_skew_normal import UnifiedSkewNormal
from botorch.utils.probability.utils import ndtr


__all__ = [
    "bvn",
    "bvnmom",
    "LinearEllipticalSliceSampler",
    "MVNXPB",
    "ndtr",
    "TruncatedMultivariateNormal",
    "UnifiedSkewNormal",
]
